from typing import (
    TYPE_CHECKING, Any, Optional, Union, overload
)
from collections.abc import Iterable, Generator, Coroutine
from typing_extensions import Literal
import os
import pathlib
import warnings
import weakref
import atexit
import logging

from opentelemetry import trace
import numpy as np
from libertem.executor.pipelined import PipelinedExecutor

from libertem.io.corrections import CorrectionSet
from libertem.executor.concurrent import ConcurrentJobExecutor
from libertem.executor.inline import InlineJobExecutor
from libertem.io.dataset import load, filetypes
from libertem.io.dataset.base import DataSet
from libertem.common.buffers import BufferWrapper
from libertem.executor.dask import DaskJobExecutor, cluster_spec
from libertem.executor.delayed import DelayedJobExecutor
from libertem.executor.integration import get_dask_integration_executor
from libertem.common.executor import JobExecutor
from libertem.common.progress import ProgressReporter
from libertem.masks import MaskFactoriesType
from libertem.analysis.raw import PickFrameAnalysis
from libertem.analysis.com import COMAnalysis
from libertem.analysis.radialfourier import RadialFourierAnalysis
from libertem.analysis.disk import DiskMaskAnalysis
from libertem.analysis.ring import RingMaskAnalysis
from libertem.analysis.sum import SumAnalysis
from libertem.analysis.point import PointMaskAnalysis
from libertem.analysis.masks import MasksAnalysis
from libertem.analysis.base import AnalysisResultSet, Analysis
from libertem.udf.base import UDFResultDict, UDF, UDFResults, UDFRunner
from libertem.udf.auto import AutoUDF
from libertem.common.async_utils import async_generator, run_agen_get_last, run_gen_get_last
from libertem.common.sparse import sparse_to_coo, to_dense
from libertem.common.exceptions import ExecutorSpecException

if TYPE_CHECKING:
    import numpy.typing as nt
    from sparse import SparseArray
    from scipy.sparse import spmatrix
    from libertem.viz.base import Live2DPlot

tracer = trace.get_tracer(__name__)
logger = logging.getLogger(__name__)

RunUDFResultType = UDFResultDict
RunUDFSyncL = list[UDFResultDict]
RunUDFAsync = Coroutine[None, None, UDFResultDict]
RunUDFAsyncL = Coroutine[None, None, list[UDFResultDict]]
ExecutorSpecType = Literal[
    'synchronous',
    'inline',
    'threads',
    'dask',
    'dask-integration',
    'dask-make-default',
    'delayed',
    'pipelined',
]
IterableRoiT = Iterable[tuple[tuple[int, ...], bool]]
RoiT = Optional[Union[np.ndarray, 'SparseArray', 'spmatrix', tuple[int, ...], IterableRoiT]]


class ResultGenerator:
    """
    Yields partial results from one or more UDFs currently running,
    with methods to control some aspects of the UDF run.
    """

    def __init__(
        self,
        task_results: Generator[UDFResults, None, None],
        runner: UDFRunner,
        result_iter,
    ):
        self._task_results = task_results
        self._runner = runner
        self._result_iter = result_iter

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._task_results)

    def close(self):
        self._task_results.close()
        self._result_iter.close()

    def update_parameters_experimental(self, parameters: list[dict[str, Any]]):
        """
        Update parameters while the UDFs are running.

        :code:`parameters` should be a list of dicts, with one dict for each
        UDF you are running.

        The dicts should only contain items for those parameters you want to
        update.
        """
        logger.debug("ResultGenerator.update_parameters_experimental: %s", parameters)
        self._result_iter.update_parameters_experimental(parameters)

    def throw(self, exc: Exception):
        return self._result_iter.throw(exc)


class ResultAsyncGenerator:
    """
    async wrapper of `ResultGenerator`.
    """

    def __init__(self, result_generator: ResultGenerator):
        from concurrent.futures import ThreadPoolExecutor
        self._result_generator = result_generator
        self._pool = ThreadPoolExecutor(max_workers=1)

    def __aiter__(self):
        return self

    async def __anext__(self):
        from .common.async_utils import MyStopIteration
        try:
            return await self._run_in_executor(self._inner_next)
        except MyStopIteration:
            raise StopAsyncIteration()

    def _inner_next(self):
        from .common.async_utils import MyStopIteration
        try:
            return next(self._result_generator)
        except StopIteration:
            raise MyStopIteration()

    async def aclose(self):
        return await self._run_in_executor(self._result_generator.close)

    async def _run_in_executor(self, f, *args):
        import asyncio
        loop = asyncio.get_event_loop()
        next_result = await loop.run_in_executor(self._pool, f, *args)
        return next_result

    async def update_parameters_experimental(self, parameters: list[dict[str, Any]]):
        """
        Update parameters while the UDFs are running.

        :code:`parameters` should be a list of dicts, with one dict for each
        UDF you are running.

        The dicts should only contain items for those parameters you want to
        update.
        """
        logger.debug("ResultGenerator.update_parameters_experimental: %s", parameters)
        return await self._run_in_executor(
            self._result_generator.update_parameters_experimental, parameters
        )

    async def athrow(self, exc: Exception):
        return await self._run_in_executor(self._result_generator.throw, exc)


RunUDFGenType = ResultGenerator
RunUDFGenTypeL = ResultGenerator
RunUDFAGenType = ResultAsyncGenerator
RunUDFAGenTypeL = ResultAsyncGenerator


class Context:
    """
    Context is the main entry point of the LiberTEM API. It contains
    methods for loading datasets, creating analyses on them and running
    them. In the background, instantiating a Context creates a suitable
    executor and spins up a local Dask cluster unless the executor is
    passed to the constructor.

    .. versionchanged:: 0.7.0
        Removed deprecated methods :code:`create_mask_job`, :code:`create_pick_job`

    Parameters
    ----------

    executor : ~libertem.common.executor.JobExecutor or None
        If None, create a local dask.distributed cluster and client using
        :meth:`~libertem.executor.dask.DaskJobExecutor.make_local` with optimal configuration
        for LiberTEM. It uses all cores and compatible GPUs
        on the local system, but is not set as default Dask scheduler to not interfere
        with other uses of Dask.

    plot_class : libertem.viz.base.Live2DPlot
        Default plot class for live plotting.
        Defaults to :class:`libertem.viz.mpl.MPLLive2DPlot`.

        .. versionadded:: 0.7.0

    Attributes
    ----------

    plot_class : libertem.viz.base.Live2DPlot
        Default plot class for live plotting.
        Defaults to :class:`libertem.viz.mpl.MPLLive2DPlot`.

        .. versionadded:: 0.7.0

    Examples
    --------

    >>> ctx = libertem.api.Context()  # doctest: +SKIP

    >>> # Create a Context using an inline executor for debugging
    >>> # See also Context.make_with() for a more convenient interface!
    >>> from libertem.executor.inline import InlineJobExecutor
    >>> debug_ctx = libertem.api.Context(executor=InlineJobExecutor())
    """

    def __init__(
        self,
        executor: Optional[JobExecutor] = None,
        plot_class: Optional['Live2DPlot'] = None,
    ):
        import traceback
        if executor is None:
            executor = self._create_local_executor()
        self.executor = executor
        self._plot_class = plot_class
        self._created_at = traceback.format_stack()
        self._register_at_exit()

    @classmethod
    def make_with(
        cls,
        executor_spec: ExecutorSpecType = 'dask',
        *,
        cpus: Optional[Union[int, Iterable[int]]] = None,
        gpus: Optional[Union[int, Iterable[int]]] = None,
        plot_class: Optional['Live2DPlot'] = None,
    ) -> 'Context':
        '''
        Create a Context with a specific kind of executor.

        .. versionadded:: 0.9.0

        This simplifies creating a :class:`Context` for a number of common executor
        choices and allows specification of the resources used to create the executor.
        For more fine-grained control of resource allocation create the
        executor manually and pass it to the :class:`Context`.

        See :ref:`executors` for general information on executors.

        .. versionchanged:: 0.12.0

            Prior to version 0.12.0, this function accepted :code:`*args, **kwargs`
            and passed them to the initializer of :class:`Context`. Given that the
            Context accepts only the :code:`plot_class` keyword-argument this was
            hard-coded into this function for backwards-compatibility, enabling
            the addition of the :code:`cpus` and :code:`gpus` parameters.

        Parameters
        ----------

        executor_spec : ExecutorSpecType, optional, by default "dask"
            A string identifier for executor variants:

            "synchronous", "inline":
                Use a single-process, single-threaded
                :class:`~libertem.executor.inline.InlineJobExecutor`
            "threads":
                Use a multi-threaded :class:`~libertem.executor.concurrent.ConcurrentJobExecutor`
            "dask":
                Create a standard :class:`~libertem.executor.dask.DaskJobExecutor` without
                considering any pre-existing Dask schedulers available on the system, similar to
                the default behaviour of :code:`Context()` called with no arguments.
            "dask-integration":
                Use a JobExecutor that is compatible with the currently active Dask scheduler.
                See :func:`~libertem.executor.integration.get_dask_integration_executor` for
                more information.
            "dask-make-default":
                Create a local :code:`dask.distributed` cluster and client
                using :meth:`~libertem.executor.dask.DaskJobExecutor.make_local`.
                The Client will be set as the default Dask scheduler and will
                persist after the LiberTEM Context closes, which is suitable for downstream
                computation using :code:`dask.distributed`.
            "delayed":
                Create a :class:`~libertem.executor.delayed.DelayedJobExecutor` which performs
                computation using `dask.delayed <https://docs.dask.org/en/stable/delayed.html>`_.
                This functionality is highly experimental at this time, see
                :ref:`delayed_udfs` for more information.
            "pipelined":
                Create a :class:`~libertem.executor.pipelined.PipelinedExecutor`,
                which is suitable for multi-process streaming live processing
                using `LiberTEM-live <https://libertem.github.io/LiberTEM-live/>`_
        cpus : int | Iterable[int], optional
            The number of CPU workers to create, where possible. The meaning of
            CPU worker depends on the type of executor created - threaded executors
            will interpret :code:`cpus` to choose the number of threads, while process-based
            executors will interpret the value to choose the number of processes to spawn.
            The iterable form of of the argument is intended contain CPU-id values to
            enable cpu-pinning where the exector and the system support it, but most
            use cases will only require an integer argument. Executors where this
            parameter does not make sense will raise an error if provided.
            When unspecified the default executor behaviour is retained.
        gpus : int | Iterable[int], optional
            Similar to :code:`cpus`, specifies the number of GPU workers to create
            where the executor chosen supports it, else raise an error. The integer
            form of the argument specifies the total number of workers to create,
            assigned according to the implementation of each executor, while
            the iterable form allows assigment of multiple workers to specific GPUs
            by repeating the id of any given GPU in the argument.
            When unspecified the default executor behaviour is retained.
        plot_class : libertem.viz.base.Live2DPlot, optional
            Plot class for live plotting, passed to :class:`Context`.

        Raises
        ------
        libertem.common.exceptions.ExecutorSpecException :
            for invalid executor choice or unsupported worker specifications

        Returns
        -------
        Instance of :class:`Context` using a new instance of the specified executor.
        '''
        # The following block is temporary until the handling of cpus/gpus args is
        # pushed onto each exector
        cpu_spec = cpus is not None
        gpu_spec = gpus is not None
        has_spec = cpu_spec or gpu_spec
        limited_execs = ('inline', 'synchronous', 'dask-integration', 'delayed')
        cannot_cpus = executor_spec in limited_execs
        if cpu_spec and cannot_cpus:
            raise ExecutorSpecException(f'Executor type {executor_spec} does not support '
                                        'specifying CPU workers at this time')
        cannot_gpus = executor_spec in limited_execs + ('threads',)
        if gpu_spec and cannot_gpus:
            raise ExecutorSpecException(f'Executor type {executor_spec} does not support '
                                        'specifying GPU workers at this time')
        # Delay import here to avoid cupy import overhead
        if has_spec and executor_spec in ('dask', 'dask-make-default', 'pipelined'):
            from libertem.utils.devices import detect
            spec_args = detect()
            if cpu_spec:
                spec_args['cpus'] = cpus
            if gpu_spec:
                if len(spec_args['cudas']) == 0 and gpus:
                    raise ExecutorSpecException('Cannot specify GPU workers as no GPUs detected')
                spec_args['cudas'] = gpus
        spec = None
        if has_spec and executor_spec in ('dask', 'dask-make-default'):
            spec = cluster_spec(**spec_args)

        executor: JobExecutor
        if executor_spec in ('synchronous', 'inline'):
            executor = InlineJobExecutor()
        elif executor_spec == 'threads':
            n_threads = cpus
            try:
                n_threads = len(n_threads)
            except TypeError:
                pass
            executor = ConcurrentJobExecutor.make_local(n_threads=n_threads)
        elif executor_spec == 'dask':
            executor = DaskJobExecutor.make_local(spec=spec)
        elif executor_spec == 'dask-integration':
            executor = get_dask_integration_executor()
        elif executor_spec == 'dask-make-default':
            executor = DaskJobExecutor.make_local(
                spec=spec,
                client_kwargs={"set_as_default": True}
            )
        elif executor_spec == 'delayed':
            executor = DelayedJobExecutor()
        elif executor_spec == 'pipelined':
            if has_spec:
                spec = PipelinedExecutor.make_spec(**spec_args)
                executor = PipelinedExecutor(spec=spec)
            else:
                executor = PipelinedExecutor.make_local()
        else:
            raise ExecutorSpecException(
                f'Argument `executor_spec` is {executor_spec}. Allowed are '
                f'"synchronous", "inline", "threads", "dask", "dask-integration",'
                f'"dask-make-default" "delayed" or "pipelined".'
            )
        return cls(executor=executor, plot_class=plot_class)

    @property
    def plot_class(self) -> type['Live2DPlot']:
        if self._plot_class is None:
            from libertem.viz.mpl import MPLLive2DPlot
            self._plot_class = MPLLive2DPlot
        return self._plot_class

    @plot_class.setter
    def plot_class(self, value: type['Live2DPlot']):
        self._plot_class = value

    def load(self, filetype: str, *args, io_backend=None, **kwargs) -> DataSet:
        """
        Load a :class:`~libertem.io.dataset.base.DataSet`. As it doesn't load
        the whole data into RAM at once, you can load and process datasets
        that are bigger than your available RAM. Using fast storage (i.e.
        SSD) is advisable.

        .. versionchanged:: 0.5.0
            Added support for filetype="auto"

        .. versionchanged:: 0.6.0
            Added support for specifying the I/O backend

        Parameters
        ----------
        filetype : str
            one of: %(types)s; or use "auto" to automatically determine filetype and parameters
        io_backend : IOBackend or None
            Use a different I/O backend for this data set
        args
            passed on to the DataSet implementation
        kwargs
            passed on to the DataSet implementation

        Returns
        -------
        DataSet : libertem.io.dataset.base.DataSet
            The loaded dataset

        Note
        ----

        Additional parameters are passed to the concrete
        :class:`~libertem.io.dataset.base.DataSet` implementation.

        Note
        ----
        See :ref:`dataset api` for format-specific documentation.

        Examples
        --------

        Load a data set from a given path, automatically determinig the type:

        >>> ds = ctx.load("auto", path="...")  # doctest: +SKIP

        To configure an alternative I/O backend, in this case configuring
        the mmap backend to enable readahead hints:

        >>> from libertem.io.dataset.base import MMapBackend
        >>> io_backend = MMapBackend(enable_readahead_hints=True)
        >>> ds = ctx.load("auto", path="...", io_backend=io_backend)  # doctest: +SKIP
        """
        # delegate to libertem.io.dataset.load:
        return load(
            filetype,
            *args,
            io_backend=io_backend,
            executor=self.executor,
            enable_async=False,
            **kwargs,
        )

    # If people run with -OO, which strips docstrings, we must not
    # try to treat load.__doc__ as `str`:
    if load.__doc__ is not None:
        load.__doc__ = load.__doc__ % {"types": ", ".join(filetypes.keys())}

    def create_mask_analysis(self, factories: MaskFactoriesType, dataset: DataSet,
            use_sparse: bool = None, mask_count: int = None, mask_dtype: np.dtype = None,
            dtype: np.dtype = None) -> MasksAnalysis:
        """
        Create a mask application analysis. Each factory function should, when
        called, return a numpy array with the same shape as frames in the
        dataset (so :code:`dataset.shape.sig`).

        This is a more high-level interface than
        :class:`~libertem.udf.masks.ApplyMasksUDF` and differs in the way the
        result is returned. With :class:`~libertem.udf.masks.ApplyMasksUDF`, it
        is a single numpy array, here we split it up for each mask we apply,
        make some default visualization available etc.

        Parameters
        ----------
        factories : Union[Callable[[], array_like], Iterable[Callable[[], array_like]]]
            Function or list of functions that take no arguments
            and create masks. The returned masks can be numpy arrays,
            scipy.sparse or sparse https://sparse.pydata.org/ matrices. The mask
            factories should not reference large objects because they can create
            significant overheads when they are pickled and unpickled. If a
            single function is specified, the first dimension is interpreted as
            the mask index.
        dataset : libertem.io.dataset.base.DataSet
            dataset to work on
        use_sparse : bool or None
            * None (default): Use sparse matrix multiplication if all factory functions return a \
            sparse mask, otherwise convert all masks to dense matrices and use dense matrix \
            multiplication
            * True: Convert all masks to sparse matrices.
            * False: Convert all masks to dense matrices.
        mask_count : int, optional
            Specify the number of masks if a single factory function is
            used so that the number of masks can be determined without calling
            the factory function.
        mask_dtype : numpy.dtype, optional
            Specify the dtype of the masks so that mask dtype can be determined without
            calling the mask factory functions. This can be used to override the
            mask dtype in the result dtype determination. As an example, setting
            this to :code:`np.float32` means that returning masks of type float64 will not switch
            the calculation and result dtype to float64 or complex128.
        dtype : numpy.dtype, optional
            Specify the dtype to do the calculation in.
            Integer dtypes are possible if the numpy casting rules allow this
            for source and mask data.

        Returns
        -------
        MasksAnalysis : libertem.analysis.base.Analysis
            When run by the Context, this Analysis generates a
            :class:`libertem.analysis.masks.MasksResultSet`.

        Examples
        --------

        >>> # Use intermediate variables instead of referencing
        >>> # large complex objects like a dataset within the
        >>> # factory function
        >>> shape = dataset.shape.sig
        >>> analysis = ctx.create_mask_analysis(
        ...     factories=[lambda: np.ones(shape)],
        ...     dataset=dataset
        ... )
        >>> result = ctx.run(analysis)
        >>> result.mask_0.raw_data.shape
        (16, 16)
        """
        return MasksAnalysis(
            dataset=dataset,
            parameters={
                "factories": factories,
                "use_sparse": use_sparse,
                "mask_count": mask_count,
                "mask_dtype": mask_dtype,
                "dtype": dtype},
        )

    def create_com_analysis(self, dataset: DataSet, cx: int = None, cy: int = None,
                            mask_radius: float = None, flip_y: bool = False,
                            mask_radius_inner: float = None,
                            scan_rotation: float = 0.0) -> COMAnalysis:
        """
        Create a center-of-mass (first moment) analysis, possibly masked.

        Parameters
        ----------
        dataset
            the dataset to work on
        cx
            reference center x value
        cy
            reference center y value
        mask_radius
            mask out intensity outside of `mask_radius` from `(cy, cx)`
        mask_radius_inner
            mask out intensity except for the ring between `mask_radius_inner` and
            `mask_radius`, centered around `(cy, cx)`

            .. versionadded:: 0.8.0
        flip_y : bool
            Flip the Y coordinate. Some detectors, namely Quantum Detectors Merlin,
            may have pixel (0, 0) at the lower left corner. This has to be corrected
            to get the sign of the y shift as well as curl and divergence right.

            .. versionadded:: 0.6.0

        scan_rotation : float
            Scan rotation in degrees.
            The optics of an electron microscope can rotate the image. Furthermore, scan
            generators may allow scanning in arbitrary directions. This means that the x and y
            coordinates of the detector image are usually not parallel to the x and y scan
            coordinates. For interpretation of center of mass shifts, however, the shift vector
            in detector coordinates has to be put in relation to the position on the sample.
            The :code:`scan_rotation` parameter can be used to rotate the detector coordinates
            to match the scan coordinate system. A positive value rotates the displacement
            vector clock-wise. That means if the detector seems rotated to the right relative
            to the scan, this value should be negative to counteract this rotation.

            .. versionadded:: 0.6.0

        Returns
        -------
        COMAnalysis : libertem.analysis.base.Analysis
            When run by the Context, this Analysis generates a
            :class:`libertem.analysis.com.COMResultSet`.
        """
        if dataset.shape.nav.dims != 2:
            raise ValueError("incompatible dataset: need two navigation dimensions")
        if dataset.shape.sig.dims != 2:
            raise ValueError("incompatible dataset: need two signal dimensions")
        loc = locals()
        parameters = {
            name: loc[name]
            for name in ['cx', 'cy', 'flip_y', 'scan_rotation']
            if loc[name] is not None
        }
        if mask_radius is not None:
            parameters['r'] = mask_radius
        if mask_radius_inner is not None:
            if mask_radius is None:
                raise ValueError(
                    "incompatible parameters: must pass both `mask_radius` and "
                    "`mask_radius_inner` for annular CoM"
                )
            parameters['ri'] = mask_radius_inner
        analysis = COMAnalysis(
            dataset=dataset, parameters=parameters
        )
        return analysis

    def create_radial_fourier_analysis(self, dataset: DataSet, cx: float = None, cy: float = None,
            ri: float = None, ro: float = None, n_bins: int = None, max_order: int = None,
            use_sparse: bool = None) -> RadialFourierAnalysis:
        """
        Create an Analysis that calculates the Fourier transform of rings around the center.

        See :ref:`radialfourier app` for details on the method!

        Parameters
        ----------
        dataset
            the dataset to work on
        cx
            center x value
        cy
            center y value
        ri
            inner radius
        ro
            outer radius
        n_bins
            number of bins
        max_order
            maximum order of calculated Fourier component

        Returns
        -------
        RadialFourierAnalysis : libertem.analysis.base.Analysis
            When run by the Context, this Analysis generates a
            :class:`libertem.analysis.radialfourier.RadialFourierResultSet`.
        """
        if dataset.shape.sig.dims != 2:
            raise ValueError("incompatible dataset: need two signal dimensions")
        loc = locals()
        parameters = {
            name: loc[name]
            for name in ['cx', 'cy', 'ri', 'ro', 'n_bins', 'max_order', 'use_sparse']
            if loc[name] is not None
        }
        analysis = RadialFourierAnalysis(
            dataset=dataset, parameters=parameters
        )
        return analysis

    def create_disk_analysis(self, dataset: DataSet, cx: int = None, cy: int = None,
                             r: int = None) -> DiskMaskAnalysis:
        """
        Create an Analysis that integrates over a disk (i.e. filled circle).

        Parameters
        ----------
        dataset
            the dataset to work on
        cx
            center x value
        cy
            center y value
        r
            radius of the disk

        Returns
        -------
        DiskMaskAnalysis : libertem.analysis.base.Analysis
            When run by the Context, this Analysis generates a
            :class:`libertem.analysis.masks.SingleMaskResultSet`.
        """
        if dataset.shape.sig.dims != 2:
            raise ValueError("incompatible dataset: need two signal dimensions")
        loc = locals()
        parameters = {name: loc[name] for name in ['cx', 'cy', 'r'] if loc[name] is not None}
        return DiskMaskAnalysis(
            dataset=dataset, parameters=parameters
        )

    def create_ring_analysis(self, dataset: DataSet, cx: int = None, cy: int = None,
                             ri: int = None, ro: int = None) -> RingMaskAnalysis:
        """
        Create an Analysis that integrates over a ring.

        Parameters
        ----------
        dataset
            the dataset to work on
        cx
            center x value
        cy
            center y value
        ri
            inner radius
        ro
            outer radius

        Returns
        -------
        RingMaskAnalysis : libertem.analysis.base.Analysis
            When run by the Context, this Analysis generates a
            :class:`libertem.analysis.masks.SingleMaskResultSet`.
        """
        if dataset.shape.sig.dims != 2:
            raise ValueError("incompatible dataset: need two signal dimensions")
        loc = locals()
        parameters = {name: loc[name] for name in ['cx', 'cy', 'ri', 'ro'] if loc[name] is not None}
        return RingMaskAnalysis(
            dataset=dataset, parameters=parameters
        )

    def create_point_analysis(self, dataset: DataSet, x: int = None,
                              y: int = None) -> PointMaskAnalysis:
        """
        Create an Analysis that selects the pixel with coords (y, x) from each frame

        Returns
        -------
        PointMaskAnalysis : libertem.analysis.base.Analysis
            When run by the Context, this Analysis generates a
            :class:`libertem.analysis.masks.SingleMaskResultSet`.
        """
        if dataset.shape.nav.dims > 2:
            raise ValueError("incompatible dataset: need at most two navigation dimensions")
        parameters = {
            'cx': x,
            'cy': y,
        }
        parameters = {
            k: v
            for k, v in parameters.items()
            if v is not None
        }
        return PointMaskAnalysis(dataset=dataset, parameters=parameters)

    def create_sum_analysis(self, dataset) -> SumAnalysis:
        """
        Create an Analysis that sums all signal elements along the navigation
        dimension, preserving the signal dimension.

        Parameters
        ----------
        dataset
            the dataset to work on

        Returns
        -------
        SumAnalysis : libertem.analysis.base.Analysis
            When run by the Context, this Analysis generates a
            :class:`libertem.analysis.sum.SumResultSet`.
        """
        return SumAnalysis(dataset=dataset, parameters={})

    def create_pick_analysis(self, dataset: DataSet, x: int, y: int = None,
                             z: int = None) -> PickFrameAnalysis:
        """
        Create an Analysis that picks a single frame / signal element from (z, y, x).
        The number of parameters must match number of navigation dimensions in the dataset,
        for example if you have a 4D dataset with two signal dimensions and two navigation
        dimensions, you need to specify x and y.

        Parameters
        ----------
        dataset
            The dataset to work on
        x
            x coordinate
        y
            y coordinate
        z
            z coordinate

        Returns
        -------
        PickFrameAnalysis : libertem.analysis.base.Analysis
            When run by the Context, this Analysis generates a
            :class:`libertem.analysis.raw.PickResultSet`.

        Examples
        --------

        >>> dataset = ctx.load(
        ...     filetype="memory",
        ...     data=np.zeros([16, 16, 16, 16, 16], dtype=np.float32),
        ...     sig_dims=2
        ... )
        >>> analysis = ctx.create_pick_analysis(dataset=dataset, x=9, y=8, z=7)
        >>> result = ctx.run(analysis)
        >>> assert result.intensity.raw_data.shape == tuple(dataset.shape.sig)
        """
        loc = locals()
        parameters = {name: loc[name] for name in ['x', 'y', 'z'] if loc[name] is not None}
        return PickFrameAnalysis(dataset=dataset, parameters=parameters)

    def run(
        self, job: Analysis,
        roi: RoiT = None,
        progress: Union[bool, ProgressReporter] = False,
        corrections: Optional[CorrectionSet] = None,
    ) -> Union[np.ndarray, AnalysisResultSet]:
        """
        Run the given :class:`~libertem.analysis.base.Analysis`
        and return the result data.

        .. versionchanged:: 0.5.0
            Added the :code:`progress` parameter

        .. versionchanged:: 0.6.0
            Added the :code:`corrections` parameter

        .. versionchanged:: 0.7.0
            Removed deprecated Job support, now only UDF-based analyses are supported

        Parameters
        ----------
        job
            the analysis to run
        roi : numpy.ndarray, sparse array or coordinate tuple(s), optional
            Boolean mask of the navigation dimension. See :ref:`udf roi`.
        progress : bool | ProgressReporter
            Show progress bar. Toggle with boolean flag or supply instance of
            :class:`libertem.common.progress.ProgressReporter` for custom
            handling of progress display.
        corrections
            Corrections to apply, i.e. dark frame substraction, applying a gain map, ...

        Returns
        -------
        result : libertem.analysis.base.AnalysisResultSet
            Running an Analysis returns a :class:`libertem.analysis.base.AnalysisResultSet`.
            See the matching :code:`create_*_analysis` function for documentation
            of the specific :code:`AnalysisResultSet` subclass or :class:`numpy.ndarray` that
            is being returned.
        """
        analysis = job  # keep the old kwarg name for backward-compat.
        if roi is None:
            roi = analysis.get_roi()
        else:
            roi = sparse_to_coo(roi, analysis.dataset.shape.nav)
        udf_results: UDFResultDict = self.run_udf(  # type:ignore[assignment]
            dataset=analysis.dataset, udf=analysis.get_udf(), roi=roi,
            corrections=corrections, progress=progress,
        )
        # Here we plot only after the computation is completed, meaning the damage should be
        # the ROI or the entire nav dimension.
        # TODO live plotting following libertem.web.jobs.JobDetailHandler.run_udf
        # Current Analysis interface possibly made obsolete by #1013, so deferred
        damage: "nt.ArrayLike"
        if roi is None:
            damage = True
        else:
            damage = to_dense(roi)
        return analysis.get_udf_results(udf_results, roi, damage=damage)

    def run_udf(
        self,
        dataset: DataSet,
        udf: Union[UDF, Iterable[UDF]],
        roi: RoiT = None,
        corrections: Optional[CorrectionSet] = None,
        progress: Union[bool, ProgressReporter] = False,
        backends=None,
        plots=None,
        sync=True,
    ) -> Union[RunUDFResultType, RunUDFSyncL, RunUDFAsync, RunUDFAsyncL]:
        """
        Run :code:`udf` on :code:`dataset`, restricted to the region of interest :code:`roi`.

        .. versionchanged:: 0.5.0
            Added the :code:`progress` parameter

        .. versionchanged:: 0.6.0
            Added the :code:`corrections` and :code:`backends` parameter

        .. versionchanged:: 0.7.0
            Added the :code:`plots` and :code:`sync` parameters,
            and the ability to run multiple UDFs on the same data in a single pass.

        Raises
        ------
        UDFRunCancelled
            Either the run was cancelled using :meth:`AsyncJobExecutor.cancel`,
            or the underlying data source was interrupted.

        Parameters
        ----------
        dataset
            The dataset to work on

        udf
            UDF instance you want to run, or a list of UDF instances

        roi : numpy.ndarray, sparse array or coordinate tuple(s), optional
            Region of interest as bool mask over the navigation axes of the dataset.
            See :ref:`udf roi`.

        progress : bool | ProgressReporter
            Show progress bar. Toggle with boolean flag or supply instance of
            :class:`libertem.common.progress.ProgressReporter` for custom
            handling of progress display.

        corrections
            Corrections to apply while running the UDF. If none are given,
            the corrections that are part of the :code:`DataSet` are used,
            if there are any. See also :ref:`corrections`.

        backends : None or iterable containing 'numpy', 'cupy' and/or 'cuda'
            Restrict the back-end to a subset of the capabilities of the UDF.
            This can be useful for testing hybrid UDFs.

        plots : None or True or List[List[Union[str, Tuple[str, Callable]]]] or List[LivePlot]
            - :code:`None`: don't plot anything (default)
            - :code:`True`: plot all 2D UDF result buffers
            - :code:`List[List[...]]`: plot the named UDF buffers. Pass a list of names or
              (name, callable) tuples for each UDF you want to plot. If the callable is specified,
              it is applied to the UDF buffer before plotting.
            - :code:`List[LivePlot]`: :class:`~libertem.viz.base.LivePlot` instance for each
              channel you want to plot

            .. versionadded:: 0.7.0

        sync : bool
            By default, `run_udf` is a synchronous method. If `sync` is set to `False`,
            it is awaitable instead.

            .. versionadded:: 0.7.0

        Returns
        -------
        dict or Tuple[dict, ...]
            Return value of the UDF containing the result buffers of
            type :class:`libertem.common.buffers.BufferWrapper`. Note that a
            :class:`~libertem.common.buffers.BufferWrapper` can be used like
            a :class:`numpy.ndarray` in many cases because it implements
            :meth:`__array__`. You can access the underlying numpy array using the
            :attr:`~libertem.common.buffers.BufferWrapper.data` property.

            If a list of UDFs was passed in, the returned type is
            a Tuple[dict[str,BufferWrapper], ...].

        Examples
        --------
        Run the :class:`~libertem.udf.sum.SumUDF` on a data set:

        >>> from libertem.udf.sum import SumUDF
        >>> result = ctx.run_udf(dataset=dataset, udf=SumUDF())
        >>> np.array(result["intensity"]).shape
        (32, 32)
        >>> # intensity is the name of the result buffer, defined in the SumUDF

        Running a UDF on a subset of data:

        >>> from libertem.udf.sumsigudf import SumSigUDF
        >>> roi = np.zeros(dataset.shape.nav, dtype=bool)
        >>> roi[0, 0] = True
        >>> result = ctx.run_udf(dataset=dataset, udf=SumSigUDF(), roi=roi)
        >>> # to get the full navigation-shaped results, with NaNs where the `roi` was False:
        >>> np.array(result["intensity"]).shape
        (16, 16)
        >>> # to only get the selected results as a flat array:
        >>> result["intensity"].raw_data.shape
        (1,)
        """
        # TODO: add a more narrow type signature - instead of a Union[...], we should
        # have overloads depending on both the type of `udf` and the `Literal[...]` value
        # of `iterate`. This was not yet added because of
        # https://github.com/python/mypy/issues/6580
        # In short, we can't have an overload `run_udf(..., plots=None, sync: Literal[True])`
        # because either we have a non-default argument after a default argument, or we have
        # `Literal[True] = ...` which overlaps with `Literal[False] = ...``
        with tracer.start_as_current_span("Context.run_udf"):
            if sync:
                return self._run_sync(
                    dataset=dataset,
                    udf=udf,
                    roi=roi,
                    corrections=corrections,
                    progress=progress,
                    backends=backends,
                    plots=plots,
                    iterate=False,
                )
            else:
                return self._run_async(
                    dataset=dataset,
                    udf=udf,
                    roi=roi,
                    corrections=corrections,
                    progress=progress,
                    backends=backends,
                    plots=plots,
                    iterate=False,
                )

    def run_udf_iter(
        self,
        dataset: DataSet,
        udf: Union[UDF, Iterable[UDF]],
        roi: RoiT = None,
        corrections: CorrectionSet = None,
        progress: Union[bool, ProgressReporter] = False,
        backends=None,
        plots=None,
        sync=True,
    ) -> Union[RunUDFGenType, RunUDFAGenType, RunUDFGenTypeL, RunUDFAGenTypeL]:
        """
        Run :code:`udf` on :code:`dataset`, restricted to the region of interest :code:`roi`.
        Yields partial results after each merge operation.

        .. versionadded:: 0.7.0

        Parameters
        ----------
        dataset
            The dataset to work on

        udf
            UDF instance you want to run, or a list of UDF instances

        roi : numpy.ndarray, sparse array or coordinate tuple(s), optional
            Region of interest as bool mask over the navigation axes of the dataset.
            See :ref:`udf roi`.

        progress : bool | ProgressReporter
            Show progress bar. Toggle with boolean flag or supply instance of
            :class:`libertem.common.progress.ProgressReporter` for custom
            handling of progress display.

        corrections
            Corrections to apply while running the UDF. If none are given,
            the corrections that are part of the :code:`DataSet` are used,
            if there are any. See also :ref:`corrections`.

        backends : None or iterable containing 'numpy', 'cupy' and/or 'cuda'
            Restrict the back-end to a subset of the capabilities of the UDF.
            This can be useful for testing hybrid UDFs.

        plots : None or True or List[List[Union[str, Tuple[str, Callable]]]] or List[LivePlot]
            - :code:`None`: don't plot anything (default)
            - :code:`True`: plot all 2D UDF result buffers
            - :code:`List[List[...]]`: plot the named UDF buffers. Pass a list of names or
              (name, callable) tuples for each UDF you want to plot. If the callable is specified,
              it is applied to the UDF buffer before plotting.
            - :code:`List[LivePlot]`: :class:`~libertem.viz.base.LivePlot` instance for each
              channel you want to plot

        sync : bool
            By default, `run_udf_iter` is a synchronous method. If `sync` is set to `False`,
            an async generator will be returned instead.

        Returns
        -------
        Generator[UDFResults]
            Generator of :class:`~libertem.udf.base.UDFResults` container objects.
            Their attribute :code:`buffers` is the list of result buffer dictionaries for the UDFs.
            Attribute :code:`damage` is a :class:`~libertem.common.buffers.BufferWrapper`
            of :code:`kind='nav'`, :code:`dtype=bool` indicating the positions
            in nav space that have been processed already.

        Examples
        --------
        Run the :class:`~libertem.udf.sum.SumUDF` on a data set:

        >>> from libertem.udf.sum import SumUDF
        >>> for result in ctx.run_udf_iter(dataset=dataset, udf=SumUDF()):
        ...     assert np.array(result.buffers[0]["intensity"]).shape == (32, 32)
        >>> np.array(result.buffers[0]["intensity"]).shape
        (32, 32)
        >>> # intensity is the name of the result buffer, defined in the SumUDF
        """
        if sync:
            return self._run_sync(
                dataset=dataset,
                udf=udf,
                roi=roi,
                corrections=corrections,
                progress=progress,
                backends=backends,
                plots=plots,
                iterate=True,
            )
        else:
            return self._run_async(
                dataset=dataset,
                udf=udf,
                roi=roi,
                corrections=corrections,
                progress=progress,
                backends=backends,
                plots=plots,
                iterate=True,
            )

    @overload
    def _run_sync(
        self,
        dataset: DataSet,
        udf: UDF,
        roi: RoiT,
        corrections: Optional[CorrectionSet],
        progress: Union[bool, ProgressReporter],
        backends: Optional[Any],
        plots: Optional[Any],
        iterate: Literal[False],
    ) -> UDFResultDict:
        ...

    @overload
    def _run_sync(
        self,
        dataset: DataSet,
        udf: UDF,
        roi: RoiT,
        corrections: Optional[CorrectionSet],
        progress: Union[bool, ProgressReporter],
        backends: Optional[Any],
        plots: Optional[Any],
        iterate: Literal[True],
    ) -> RunUDFGenType: ...

    @overload
    def _run_sync(
        self,
        dataset: DataSet,
        udf: Iterable[UDF],
        roi: RoiT,
        corrections: Optional[CorrectionSet],
        progress: Union[bool, ProgressReporter],
        backends: Optional[Any],
        plots: Optional[Any],
        iterate: Literal[False],
    ) -> RunUDFSyncL: ...

    @overload
    def _run_sync(
        self,
        dataset: DataSet,
        udf: Iterable[UDF],
        roi: RoiT,
        corrections: Optional[CorrectionSet],
        progress: Union[bool, ProgressReporter],
        backends: Optional[Any],
        plots: Optional[Any],
        iterate: Literal[True],
    ) -> RunUDFGenTypeL: ...

    @overload
    def _run_sync(
        self,
        dataset: DataSet,
        udf: Union[Iterable[UDF], UDF],
        roi: RoiT,
        corrections: Optional[CorrectionSet],
        progress: Union[bool, ProgressReporter],
        backends: Optional[Any],
        plots: Optional[Any],
        iterate: Literal[True],
    ) -> Union[RunUDFGenType, RunUDFGenTypeL]:
        ...

    @overload
    def _run_sync(
        self,
        dataset: DataSet,
        udf: Union[Iterable[UDF], UDF],
        roi: RoiT,
        corrections: Optional[CorrectionSet],
        progress: Union[bool, ProgressReporter],
        backends: Optional[Any],
        plots: Optional[Any],
        iterate: Literal[False],
    ) -> Union[UDFResultDict, RunUDFSyncL]:
        ...

    @overload
    def _run_sync(
        self,
        dataset: DataSet,
        udf: Union[Iterable[UDF], UDF],
        roi: RoiT,
        corrections: Optional[CorrectionSet],
        progress: Union[bool, ProgressReporter],
        backends: Optional[Any],
        plots: Optional[Any],
        iterate: bool,
    ) -> Union[UDFResultDict, RunUDFSyncL, RunUDFGenType, RunUDFGenTypeL]: ...

    def _run_sync(
        self,
        dataset: DataSet,
        udf: Union[UDF, Iterable[UDF]],
        roi: RoiT,
        corrections: Optional[CorrectionSet],
        progress: Union[bool, ProgressReporter],
        backends,
        plots,
        iterate: bool,
    ):
        """
        Run the given UDF(s), either returning the final result (when
        :code:`iterate=False` is given), or a generator that yields partial results.
        """
        enable_plotting = bool(plots)

        udf_is_list = isinstance(udf, Iterable)
        if not isinstance(udf, Iterable):  # duplicate, because mypy.
            udfs = [udf]
        else:
            udfs = list(udf)

        if len(udfs) == 0:
            raise ValueError("empty list of UDFs - nothing to do!")

        if enable_plotting:
            with tracer.start_as_current_span("prepare_plots"):
                plots = self._prepare_plots(udfs, dataset, roi, plots)

        if corrections is None:
            corrections = dataset.get_correction_data()

        if roi is not None:
            roi = sparse_to_coo(roi, dataset.shape.nav)
            if roi.dtype is not np.dtype(bool):
                warnings.warn(f"ROI dtype is {roi.dtype}, expected bool. Attempting cast to bool.")
                roi = roi.astype(bool)

        progress_reporter: Optional[ProgressReporter]
        if isinstance(progress, ProgressReporter):
            progress_reporter = progress
            progress = True
        else:
            progress_reporter = None

        def _run_sync_wrap() -> ResultGenerator:
            runner_cls = self.executor.get_udf_runner()
            runner = runner_cls(
                udfs,
                progress_reporter=progress_reporter,
            )
            result_iter = runner.run_for_dataset_sync(
                dataset=dataset,
                executor=self.executor,
                roi=roi,
                progress=progress,
                corrections=corrections,
                backends=backends,
                iterate=(iterate or enable_plotting)
            )

            def _inner():
                for udf_results in result_iter:
                    yield udf_results
                    if enable_plotting:
                        self._update_plots(
                            plots, udfs, udf_results.buffers, udf_results.damage.data, force=False
                        )
                if enable_plotting:
                    self._update_plots(
                        plots, udfs, udf_results.buffers, udf_results.damage.data, force=True
                    )
            return ResultGenerator(_inner(), runner=runner, result_iter=result_iter)

        if iterate:
            return _run_sync_wrap()
        else:
            udf_results = run_gen_get_last(_run_sync_wrap())
            if udf_is_list:
                return udf_results.buffers
            else:
                return udf_results.buffers[0]

    @overload
    def _run_async(
        self,
        dataset: DataSet,
        udf: UDF,
        roi: RoiT,
        corrections: Optional[CorrectionSet],
        progress: Union[bool, ProgressReporter],
        backends,
        plots,
        iterate: Literal[False],
    ) -> RunUDFAsync: ...

    @overload
    def _run_async(
        self,
        dataset: DataSet,
        udf: Iterable[UDF],
        roi: RoiT,
        corrections: Optional[CorrectionSet],
        progress: Union[bool, ProgressReporter],
        backends,
        plots,
        iterate: Literal[False],
    ) -> RunUDFAsyncL: ...

    @overload
    def _run_async(
        self,
        dataset: DataSet,
        udf: UDF,
        roi: RoiT,
        corrections: Optional[CorrectionSet],
        progress: Union[bool, ProgressReporter],
        backends,
        plots,
        iterate: Literal[True],
    ) -> RunUDFAGenType: ...

    @overload
    def _run_async(
        self,
        dataset: DataSet,
        udf: Iterable[UDF],
        roi: RoiT,
        corrections: Optional[CorrectionSet],
        progress: Union[bool, ProgressReporter],
        backends,
        plots,
        iterate: Literal[True],
    ) -> RunUDFAGenTypeL: ...

    @overload
    def _run_async(
        self,
        dataset: DataSet,
        udf: Union[UDF, Iterable[UDF]],
        roi: RoiT,
        corrections: Optional[CorrectionSet],
        progress: Union[bool, ProgressReporter],
        backends,
        plots,
        iterate: Literal[True],
    ) -> Union[RunUDFAGenTypeL, RunUDFAGenType]: ...

    @overload
    def _run_async(
        self,
        dataset: DataSet,
        udf: Union[UDF, Iterable[UDF]],
        roi: RoiT,
        corrections: Optional[CorrectionSet],
        progress: Union[bool, ProgressReporter],
        backends,
        plots,
        iterate: Literal[False],
    ) -> Union[RunUDFAsync, RunUDFAsyncL]: ...

    @overload
    def _run_async(
        self,
        dataset: DataSet,
        udf: Union[Iterable[UDF], UDF],
        roi: RoiT,
        corrections: Optional[CorrectionSet],
        progress: Union[bool, ProgressReporter],
        backends,
        plots,
        iterate: bool,
    ) -> Union[RunUDFAsync, RunUDFAsyncL, RunUDFAGenType, RunUDFAGenTypeL]: ...

    def _run_async(
        self,
        dataset: DataSet,
        udf: Union[UDF, Iterable[UDF]],
        roi: RoiT,
        corrections: Optional[CorrectionSet],
        progress: Union[bool, ProgressReporter],
        backends,
        plots,
        iterate: bool,
    ):
        """
        Wraps :code:`_run_sync` into an asynchronous generator,
        and either returns the generator itself, or the end result.
        """
        sync_generator = self._run_sync(
            dataset=dataset,
            udf=udf,
            roi=roi,
            corrections=corrections,
            progress=progress,
            backends=backends,
            plots=plots,
            iterate=True,
        )

        async def _run_async_wrap() -> UDFResultDict:
            udf_results = await run_agen_get_last(async_generator(sync_generator))
            return udf_results.buffers[0]

        async def _run_async_wrap_l() -> list[UDFResultDict]:
            udf_results = await run_agen_get_last(async_generator(sync_generator))
            return udf_results.buffers

        if iterate:
            return ResultAsyncGenerator(result_generator=sync_generator)
        else:
            if isinstance(udf, Iterable):
                return _run_async_wrap_l()
            else:
                return _run_async_wrap()

    def _get_default_plot_chans(self, buffers):
        from libertem.viz import get_plottable_2D_channels
        return [
            get_plottable_2D_channels(bufferset)
            for bufferset in buffers
        ]

    def _prepare_plots(
        self,
        udfs: list[UDF],
        dataset: DataSet,
        roi: RoiT,
        plots,
    ) -> list['Live2DPlot']:
        runner_cls = self.executor.get_udf_runner()
        dry_results = runner_cls.dry_run(udfs, dataset, roi)

        # cases to consider:
        # 1) plots is `True`: default plots of all eligible channels
        # 2) plots is List[List[str]] or List[List[(str, callable)]]: set channels from `plots`
        # 3) plots is List[LivePlot]: use customized plots as they are

        channels = None

        # 1) plots is `True`: default plots of all eligible channels
        if plots is True:
            channels = self._get_default_plot_chans(dry_results.buffers)
            for idx, udf in enumerate(udfs):
                if len(channels[idx]) == 0:
                    warnings.warn(
                        f"No plottable channels found for UDF "
                        f"#{idx}: {udf.__class__.__name__}, not plotting."
                    )
        # 2) plots is List[List[str]] or List[List[(str, callable)]]: set channels from `plots`
        elif (isinstance(plots, (list, tuple))
                and all(isinstance(p, (list, tuple)) for p in plots)
                and all(all(isinstance(pp, (str, list, tuple)) for pp in p) for p in plots)):
            channels = plots
        # 3) plots is probably List[LivePlot]: use customized plots as they are
        else:
            return plots

        plots = []
        for idx, (udf, udf_channels) in enumerate(zip(udfs, channels)):
            for channel in udf_channels:
                p0 = self.plot_class(
                    dataset,
                    udf=udf,
                    roi=roi,
                    channel=channel,
                    # Create an UDFResult from this single UDF
                    udfresult=UDFResults(
                        (dry_results.buffers[idx],),
                        dry_results.damage
                    ),
                )
                p0.display()
                plots.append(p0)
        return plots

    def _update_plots(
        self,
        plots: list['Live2DPlot'],
        udfs: list[UDF],
        udf_results: tuple[dict[str, BufferWrapper], ...],
        damage: np.ndarray,
        force: bool = False,
    ):
        for plot in plots:
            udf = plot.get_udf()
            udf_index = udfs.index(udf)
            plot.new_data(udf_results[udf_index], damage, force=force)

    def display(self, dataset: DataSet, udf: UDF, roi: RoiT = None):
        """
        Show information about the UDF in combination with the given DataSet.
        """
        import html

        class _UDFInfo:
            def __init__(self, title, buffers):
                self.title = title
                self.buffers = buffers

            def _repr_html_(self):
                def _e(obj):
                    return html.escape(str(obj))

                rows = [
                    "<tr>"
                    f"<td>{_e(key)}</td>"
                    f"<td>{_e(buf.kind)}</td>"
                    f"<td>{_e(buf.extra_shape)}</td>"
                    f"<td>{_e(buf.shape)}</td>"
                    f"<td>{_e(buf.dtype)}</td>"
                    "</tr>"
                    for key, buf in self.buffers.items()
                    if buf.use != "private"
                ]
                rows = "\n".join(rows)
                general = f"""
                <table>
                    <tbody>
                        <tr>
                            <th>Processing method</th>
                            <td>{_e(udf.get_method())}</td>
                        </tr>
                        <tr>
                            <th>Compute Backends</th>
                            <td>{_e(" ,".join(udf.get_backends()))}</td>
                        </tr>
                        <tr>
                            <th>Preferred input dtype</th>
                            <td>{_e(np.dtype(udf.get_preferred_input_dtype()))}</td>
                        </tr>
                    </tbody>
                </table>
                """
                return f"""
                <h2>{_e(self.title)}</h2>
                <h3>General</h3>
                {general}
                <h3>Result types</h3>
                <p>Note: these may vary with different data sets</p>
                <table>
                    <thead>
                        <th>Name</th>
                        <th>Kind</th>
                        <th>Extra Shape</th>
                        <th>Concrete Shape</th>
                        <th>dtype</th>
                    </thead>
                    <tbody>
                    {rows}
                    </tbody>
                </table>
                """
        runner_cls = self.executor.get_udf_runner()

        if roi is not None:
            roi = sparse_to_coo(roi, dataset.shape.nav)

        return _UDFInfo(
            title=udf.__class__.__name__,
            buffers=runner_cls.inspect_udf(udf, dataset, roi),
        )

    def map(self, dataset: DataSet, f, roi: RoiT = None,
            progress: Union[bool, ProgressReporter] = False,
            corrections: CorrectionSet = None,
            backends=None) -> BufferWrapper:
        '''
        Create an :class:`AutoUDF` with function :meth:`f` and run it on :code:`dataset`

        .. versionchanged:: 0.5.0
            Added the :code:`progress` parameter

        .. versionchanged:: 0.6.0
            Added the :code:`corrections` and :code:`backends` parameter

        Parameters
        ----------

        dataset:
            The dataset to work on
        f:
            Function that accepts a frame as the only parameter. It should return a strongly
            reduced output compared to the size of a frame.
        roi : numpy.ndarray, sparse array or coordinate tuple(s), optional
            Region of interest as bool mask over the navigation axes of the dataset.
            See :ref:`udf roi`.
        progress : bool | ProgressReporter
            Show progress bar. Toggle with boolean flag or supply instance of
            :class:`libertem.common.progress.ProgressReporter` for custom
            handling of progress display.
        corrections
            Corrections to apply while running the function. If none are given,
            the corrections that are part of the :code:`DataSet` are used,
            if there are any. See also :ref:`corrections`.
        backends : None or iterable containing 'numpy', 'cupy' and/or 'cuda'
            Restrict the back-end to a subset of the capabilities of the UDF.
            This can be useful for testing hybrid UDFs.

        Returns
        -------

        BufferWrapper : libertem.common.buffers.BufferWrapper
            The result of the UDF. Access the underlying numpy array using the
            :attr:`~libertem.common.buffers.BufferWrapper.data` property.
            Shape and dtype is inferred automatically from :code:`f`.
        '''
        udf = AutoUDF(f=f)
        results: UDFResultDict = self.run_udf(  # type:ignore[assignment]
            dataset=dataset,
            udf=udf,
            roi=roi,
            progress=progress,
            corrections=corrections,
            backends=backends,
        )
        return results['result']

    def _create_local_executor(self):
        return DaskJobExecutor.make_local()

    def close(self):
        self.executor.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def _register_at_exit(self):
        """
        Register at-exit handler, to make sure the executor is closed
        """
        weak_ctx = weakref.ref(self)

        def _exit():
            if weak_ctx() is None:
                return
            try:
                weak_ctx().close()
            except Exception:  # pragma: no cover
                # can't be sure that logging is still alive,
                # so we have to fall back to a normal print here:
                print("\n\n\n\n\n")
                print("Exception in atexit handler for Context created here:")
                print("".join(weak_ctx()._created_at))
                print("\n\n\n\n\n")
                raise

        atexit.register(_exit)

    def export_dataset(
        self,
        dataset: DataSet,
        *,
        path: os.PathLike,
        progress: bool = False,
        overwrite: bool = False,
    ):
        """
        Export the dataset to another format on disk

        At this time can only convert to numpy *.npy* format,
        but future extensions are possible.

        The written data will have any reshaping / sync_offset
        properties of the DataSet effectively baked into
        the new file.

        .. note::
            All workers used by the Context must have access to the
            same file path, over a network file system if necessary,
            in order to correctly save data to the file.

        .. versionadded:: 0.12.0

        Parameters
        ----------
        dataset : lt.Dataset
            The dataset to save to disk
        path : os.PathLike
            The file path to export the data to, will
            raise ValueError if the suffix is unrecognized
            (currently supports only .npy)
        progress : bool, optional
            Whether to display a progress bar for the export,
            by default False
        overwrite : bool , optional
            If the save path already exists, raise FileExistsError unless
            overwrite is True, by default False

        Raises
        ------
        ValueError
            If the path suffix is not supported
        FileExistsError
            If overwrite is True and the save path exists
        """
        path = pathlib.Path(path)
        if path.suffix != '.npy':
            raise ValueError(
                f'Unrecognized file extension {path.suffix} '
                'only .npy is currently supported.'
            )

        if not overwrite and path.is_file():
            raise FileExistsError(
                f'Cannot export dataset to existing path {path} .'
                'Use overwrite=True to force export.'
            )

        from libertem.udf.record import RecordUDF
        udf = RecordUDF(path)
        self.run_udf(dataset, udf, progress=progress)



import math

import pytest
import numpy as np
import numba

from libertem.io.dataset.k2is import decode_uint12_le


@numba.njit
def decode_uint12_le_ref(inp, out):
    """
    decode bytes from bytestring ``inp`` as 12 bit into ``out``
    """
    o = 0
    for i in range(0, len(inp), 3):
        s = inp[i:i + 3]
        a = s[0] | (s[1] & 0x0F) << 8
        b = (s[1] & 0xF0) >> 4 | s[2] << 4
        out[o] = a
        out[o + 1] = b
        o += 2


def encode_12_little_little(inp, out):
    bytebits = 8
    bits = 12
    out = out.view(np.uint8)
    encoded_type = np.dtype('<u4')
    decoded_words = len(inp)
    assert len(out)*bytebits >= decoded_words*bits

    encoded_bytes = encoded_type.itemsize

    # how many words of encoded_type an encoded block
    encoded_stride = 3
    # how many words of "bits" bits in an encoded block
    decoded_stride = 8

    mask = 0xfff

    # How many blocks are encoded in one loop
    middle_loops = 8
    # How many decoded words processed in one middle loop
    middle_decoded_words = middle_loops * decoded_stride
    middle_encoded_words = middle_loops * encoded_stride
    middle_encoded_bytes = middle_encoded_words * encoded_bytes

    blocks = decoded_words // middle_decoded_words

    rest = decoded_words % middle_decoded_words

    def decoded_start_index(block):
        return block*middle_decoded_words

    def encoded_start_index(block):
        return block*middle_encoded_words

    def decoded_loopbuffer_index(middle):
        return middle*decoded_stride

    def encoded_loopbuffer_index(middle):
        return middle*encoded_stride

    def char_start_index(block):
        return block*middle_encoded_words*encoded_bytes

    def encode_pair(k, m):
        a = np.uint8(k & 0xff)
        b = np.uint8(((k & 0xf00) >> 8) | ((m & 0xf) << 4))
        c = np.uint8((m & 0xff0) >> 4)
        return (a, b, c)

    loopbuffer_in = np.zeros(middle_decoded_words, dtype=inp.dtype)
    loopbuffer_out = np.zeros(middle_encoded_words, dtype=encoded_type)
    loopbuffer_out_chars = loopbuffer_out.view(np.uint8)

    for block in range(blocks):
        decoded_start = decoded_start_index(block)
        for i in range(middle_decoded_words):
            loopbuffer_in[i] = inp[decoded_start + i]

        for middle in range(middle_loops):
            decoded_base = decoded_loopbuffer_index(middle)
            encoded_base = encoded_loopbuffer_index(middle)
            loopbuffer_out[encoded_base + 0] = (
                (loopbuffer_in[decoded_base + 0] & mask)
                | ((loopbuffer_in[decoded_base + 1] & mask) << 12)
                | ((loopbuffer_in[decoded_base + 2] & mask) << 24)
            )

            loopbuffer_out[encoded_base + 1] = (
                ((loopbuffer_in[decoded_base + 2] & mask) >> 8)
                | ((loopbuffer_in[decoded_base + 3] & mask) << 4)
                | ((loopbuffer_in[decoded_base + 4] & mask) << 16)
                | ((loopbuffer_in[decoded_base + 5] & mask) << 28)
            )

            loopbuffer_out[encoded_base + 2] = (
                ((loopbuffer_in[decoded_base + 5] & mask) >> 4)
                | ((loopbuffer_in[decoded_base + 6] & mask) << 8)
                | ((loopbuffer_in[decoded_base + 7] & mask) << 20)
            )
        char_start = char_start_index(block)
        for i in range(middle_encoded_bytes):
            out[char_start + i] = loopbuffer_out_chars[i]
    decoded_remainder_offset = decoded_start_index(blocks)
    encoded_remainder_offset = char_start_index(blocks)
    if (rest > 1):
        for r in range(0, rest - 1, 2):
            k = inp[decoded_remainder_offset + r]
            m = inp[decoded_remainder_offset + r + 1]
            (a, b, c) = encode_pair(k, m)
            out[encoded_remainder_offset] = a
            out[encoded_remainder_offset + 1] = b
            out[encoded_remainder_offset + 2] = c
            encoded_remainder_offset += 3
    if (rest % 2):
        k = inp[decoded_remainder_offset + rest - 1]
        m = 0
        (a, b, c) = encode_pair(k, m)
        out[encoded_remainder_offset] = a
        out[encoded_remainder_offset + 1] = b


@pytest.mark.with_numba
def test_encode_decode_uint12_ref():
    mult = 45
    inp = np.arange(2*mult, dtype=np.uint16)
    out = np.zeros(int(math.ceil(len(inp) * 12 / 8)), dtype=np.uint8)
    result = np.zeros(int(math.ceil(len(out) * 2 / 3)), dtype=np.uint16)

    print(list(map(hex, inp)))
    encode_12_little_little(inp, out)

    print(list(map(hex, out.view(np.uint8))))
    decode_uint12_le_ref(inp=out, out=result)

    assert np.allclose(inp, result)


@pytest.mark.with_numba
def test_encode_decode_uint12():
    mult = 45
    inp = np.arange(2*mult, dtype=np.uint16)
    out = np.zeros(int(math.ceil(len(inp) * 12 / 8)), dtype=np.uint8)
    result = np.zeros(int(math.ceil(len(out) * 2 / 3)), dtype=np.uint16)
    result_ref = np.zeros(int(math.ceil(len(out) * 2 / 3)), dtype=np.uint16)

    print(list(map(hex, inp)))

    encode_12_little_little(inp, out)
    print(list(map(hex, out.view(np.uint8))))

    decode_uint12_le(inp=out, out=result)
    assert np.allclose(inp, result)

    decode_uint12_le_ref(inp=out, out=result_ref)
    assert np.allclose(result_ref, result)



import os
import sys
import json
import random
import glob

import pytest
import numpy as np
import cloudpickle

from libertem.io.dataset.frms6 import (
    FRMS6DataSet, _map_y, FRMS6Decoder, get_filenames
)
from libertem.analysis.raw import PickFrameAnalysis
from libertem.analysis.sum import SumAnalysis
from libertem.udf.sumsigudf import SumSigUDF
from libertem.io.dataset.base import TilingScheme, BufferedBackend, MMapBackend, DirectBackend
from libertem.common import Shape
from libertem.common.buffers import reshaped_view
from libertem.udf.raw import PickUDF

from utils import (dataset_correction_verification, get_testdata_path, ValidationUDF,
    FakeBackend, roi_as_sparse)

try:
    import stemtool
except ModuleNotFoundError:
    stemtool = None


FRMS6_TESTDATA_PATH = os.path.join(get_testdata_path(), 'frms6', 'C16_15_24_151203_019.hdr')
HAVE_FRMS6_TESTDATA = os.path.exists(FRMS6_TESTDATA_PATH)

needsdata = pytest.mark.skipif(not HAVE_FRMS6_TESTDATA, reason="need frms6 testdata")  # NOQA


@pytest.fixture
def default_frms6(lt_ctx):
    ds = lt_ctx.load(
        "frms6",
        path=FRMS6_TESTDATA_PATH,
        io_backend=MMapBackend(),
    )
    return ds


@pytest.fixture
def buffered_frms6(lt_ctx):
    buffered = BufferedBackend()
    return lt_ctx.load(
        "frms6",
        path=str(FRMS6_TESTDATA_PATH),
        io_backend=buffered,
    )


@pytest.fixture
def direct_frms6(lt_ctx):
    direct = DirectBackend()
    return lt_ctx.load(
        "frms6",
        path=str(FRMS6_TESTDATA_PATH),
        io_backend=direct,
    )


@pytest.fixture
def default_frms6_uncorr(lt_ctx):
    ds = FRMS6DataSet(path=FRMS6_TESTDATA_PATH, enable_offset_correction=False)
    ds = ds.initialize(lt_ctx.executor)
    return ds


def _read_block(fname, raw_shape, start, stop):
    import stemtool  # avoid importing on top level
    return stemtool.util.pnccd.Frms6Reader.readData(
        fname,
        image_range=(start, stop),
        pixels_x=raw_shape[0],
        pixels_y=raw_shape[1]
    )


def _unfold_block(block, unbinning_factor=4):
    return np.moveaxis(  # undo the transpose
        np.repeat(  # unbinning in x direction
            # invert lower half and attach right of upper half
            # The detector consists of two chips that are arranged head-to-head
            # The outputs of the two chips are just concatenated in the file, while LiberTEM
            # re-assembles the data taking the spatial relation into account
            np.concatenate((block[:264], np.flip(block[264:], axis=(0, 1,))), axis=1),
            unbinning_factor, axis=1  # repeat options unbinning x
        ),
        (0, 1, 2), (2, 1, 0)  # moveaxis options undo transpose
    )


@pytest.fixture(scope='module')
def default_frms6_raw(tmpdir_factory):
    fn = tmpdir_factory.mktemp("data").join("frms6.raw")
    # we use a memory mapped file to make this work
    # on machines that can't hold the full dataset in memory
    data = np.memmap(str(fn), mode='w+', shape=(256, 256, 264, 264), dtype='uint16')
    view = reshaped_view(data, (256*256, 264, 264))
    root, ext = os.path.splitext(FRMS6_TESTDATA_PATH)
    files = list(sorted(glob.glob(root + '*.frms6')))
    blocksize = 15
    offset = 0
    # we skip the first file, it contains a zero reference
    for f in files[1:]:
        raw_shape = stemtool.util.pnccd.Frms6Reader.getDataShape(f)
        frame_count = raw_shape[-1]
        # We go blockwise to reduce memory consumption
        for start in range(0, frame_count, blocksize):
            stop = min(start+blocksize, frame_count)
            block = _read_block(f, raw_shape, start, stop)
            view[offset + start:offset + stop] = _unfold_block(block)
        offset += frame_count

    yield data

    # Might be needed for windows, so we don't keep any views, mmaps or open
    # file handles around:
    del view
    data._mmap.close()
    del data
    os.unlink(fn)


@pytest.fixture(scope='module')
def default_frms6_darkref():
    import stemtool  # avoid importing on top level
    root, ext = os.path.splitext(FRMS6_TESTDATA_PATH)
    # we grab the first file
    f = list(sorted(glob.glob(root + '*.frms6')))[0]
    blocksize = 15
    raw_shape = stemtool.util.pnccd.Frms6Reader.getDataShape(f)
    frame_count = raw_shape[-1]
    # We go blockwise to reduce memory consumption
    result = np.zeros((264, 264), dtype=np.float32)
    for start in range(0, frame_count, blocksize):
        stop = min(start+blocksize, frame_count)
        block = _read_block(f, raw_shape, start, stop)
        result += _unfold_block(block).sum(axis=0)
    return result / frame_count


def test_glob(tmp_path, naughty_filename):
    naughty_dir = tmp_path / naughty_filename
    naughty_dir.mkdir()

    naughty_hdr = naughty_dir / (naughty_filename + '.hdr')
    naughty_hdr.touch()

    naughty_frms6_0 = naughty_dir / (naughty_filename + '_000.frms6')
    naughty_frms6_1 = naughty_dir / (naughty_filename + '_001.frms6')
    naughty_frms6_0.touch()
    naughty_frms6_1.touch()

    filenames_1 = set(get_filenames(naughty_hdr))
    filenames_2 = set(get_filenames(naughty_frms6_0))
    print(filenames_1)
    print(filenames_2)
    target = {str(naughty_frms6_0), str(naughty_frms6_1)}
    assert filenames_1 == target
    assert filenames_2 == target


@needsdata
def test_simple_open(default_frms6):
    assert tuple(default_frms6.shape) == (256, 256, 264, 264)


@needsdata
def test_auto_open_corrections_kwargs(lt_ctx):
    ds_corr = lt_ctx.load(
        'auto', path=FRMS6_TESTDATA_PATH, enable_offset_correction=True, nav_shape=(2, 3)
    )
    assert not np.allclose(ds_corr.get_correction_data().get_dark_frame(), 0)
    assert tuple(ds_corr.shape.nav) == (2, 3)
    assert isinstance(ds_corr, FRMS6DataSet)

    ds = lt_ctx.load(
        'auto', path=FRMS6_TESTDATA_PATH, enable_offset_correction=False, nav_shape=(2, 3)
    )
    assert not ds.get_correction_data().have_corrections()
    assert tuple(ds.shape.nav) == (2, 3)
    assert isinstance(ds, FRMS6DataSet)


@needsdata
def test_auto_open_corrections_posargs(lt_ctx):
    ds_corr = lt_ctx.load('auto', FRMS6_TESTDATA_PATH, True, None, None, (2, 3))
    assert not np.allclose(ds_corr.get_correction_data().get_dark_frame(), 0)
    assert tuple(ds_corr.shape.nav) == (2, 3)
    assert isinstance(ds_corr, FRMS6DataSet)

    ds = lt_ctx.load('auto', FRMS6_TESTDATA_PATH, False, None, None, (2, 3))
    assert not ds.get_correction_data().have_corrections()
    assert tuple(ds.shape.nav) == (2, 3)
    assert isinstance(ds, FRMS6DataSet)

    ds_corr = lt_ctx.load('frms6', FRMS6_TESTDATA_PATH, True, None, None, (2, 3))
    assert not np.allclose(ds_corr.get_correction_data().get_dark_frame(), 0)
    assert tuple(ds_corr.shape.nav) == (2, 3)
    assert isinstance(ds_corr, FRMS6DataSet)

    ds = lt_ctx.load('frms6', FRMS6_TESTDATA_PATH, False, None, None, (2, 3))
    assert not ds.get_correction_data().have_corrections()
    assert tuple(ds.shape.nav) == (2, 3)
    assert isinstance(ds, FRMS6DataSet)


@needsdata
def test_auto_uses_correct_backend(lt_ctx):
    with pytest.raises(RuntimeError):
        ds = lt_ctx.load(
            "auto",
            path=FRMS6_TESTDATA_PATH,
            io_backend=FakeBackend(),
        )
        lt_ctx.run_udf(
            dataset=ds,
            udf=SumSigUDF(),
        )


@needsdata
def test_detetct(lt_ctx):
    assert FRMS6DataSet.detect_params(
        FRMS6_TESTDATA_PATH, lt_ctx.executor
    )["parameters"] is not False


@needsdata
def test_check_valid(default_frms6):
    default_frms6.check_valid()


@needsdata
def test_sum_analysis(default_frms6, local_cluster_ctx):
    roi = {
        "shape": "disk",
        "cx": 5,
        "cy": 6,
        "r": 7,
    }
    analysis = SumAnalysis(dataset=default_frms6, parameters={
        "roi": roi,
    })
    # not checking result yet, just making sure it doesn't crash:
    local_cluster_ctx.run(analysis)


@needsdata
def test_pick_analysis(default_frms6, lt_ctx):
    analysis = PickFrameAnalysis(dataset=default_frms6, parameters={"x": 16, "y": 16})
    results = lt_ctx.run(analysis)
    assert results[0].raw_data.shape == (264, 264)


@needsdata
@pytest.mark.parametrize(
    # Default is too large for test without ROI
    "with_roi", (True, )
)
def test_correction(default_frms6, lt_ctx, with_roi):
    ds = default_frms6

    if with_roi:
        roi = np.zeros(ds.shape.nav, dtype=bool)
        roi[:1] = True
    else:
        roi = None

    dataset_correction_verification(ds=ds, roi=roi, lt_ctx=lt_ctx)


@needsdata
@pytest.mark.skipif(stemtool is None, reason="No stemtool found")
def test_comparison(default_frms6_uncorr, default_frms6_raw, lt_ctx_fast):
    udf = ValidationUDF(
        reference=reshaped_view(default_frms6_raw, (-1, *tuple(default_frms6_uncorr.shape.sig))),
        validation_function=lambda a, b: np.all(a == b),
    )
    lt_ctx_fast.run_udf(udf=udf, dataset=default_frms6_uncorr)


@needsdata
@pytest.mark.skipif(stemtool is None, reason="No stemtool found")
def test_comparison_roi(default_frms6_uncorr, default_frms6_raw, lt_ctx_fast):
    roi = np.random.choice(
        [True, False],
        size=tuple(default_frms6_uncorr.shape.nav),
        p=[0.001, 0.999]
    )
    udf = ValidationUDF(
        reference=default_frms6_raw[roi],
        validation_function=lambda a, b: np.all(a == b),
    )
    lt_ctx_fast.run_udf(udf=udf, dataset=default_frms6_uncorr, roi=roi)


@needsdata
@pytest.mark.skipif(stemtool is None, reason="No stemtool found")
def test_comparison_darkref(default_frms6, default_frms6_darkref):
    corr_ds = default_frms6.get_correction_data().get_dark_frame()
    print(np.max(corr_ds - default_frms6_darkref))
    assert np.allclose(corr_ds, default_frms6_darkref)


@needsdata
def test_pickle_is_small(default_frms6):
    pickled = cloudpickle.dumps(default_frms6)
    cloudpickle.loads(pickled)

    # because of the dark frame stuff, the dataset is actually quite large:
    assert len(pickled) < 300 * 1024


@needsdata
def test_cache_key_json_serializable(default_frms6):
    json.dumps(default_frms6.get_cache_key())


@needsdata
@pytest.mark.dist
def test_dist_process(default_frms6, dist_ctx):
    roi = {
        "shape": "disk",
        "cx": 5,
        "cy": 6,
        "r": 7,
    }
    analysis = SumAnalysis(dataset=default_frms6, parameters={"roi": roi})
    dist_ctx.run(analysis)


@needsdata
@pytest.mark.dist
def test_initialize(default_frms6, dist_ctx):
    assert default_frms6._filenames is not None
    assert default_frms6._hdr_info is not None
    assert default_frms6._hdr_info is not None

# TODO: gain map tests
# TODO: test load request message
# TODO: test error conditions


def test_map_y():
    assert _map_y(y=0, xs=264, binning=4, num_rows=264) == (0, 0)
    assert _map_y(y=32, xs=264, binning=4, num_rows=264) == (32, 0)
    assert _map_y(y=33, xs=264, binning=4, num_rows=264) == (32, 264)
    assert _map_y(y=65, xs=264, binning=4, num_rows=264) == (0, 264)

    assert _map_y(y=0, xs=264, binning=1, num_rows=264) == (0, 0)
    assert _map_y(y=32, xs=264, binning=1, num_rows=264) == (32, 0)
    assert _map_y(y=33, xs=264, binning=1, num_rows=264) == (33, 0)
    assert _map_y(y=65, xs=264, binning=1, num_rows=264) == (65, 0)
    assert _map_y(y=131, xs=264, binning=1, num_rows=264) == (131, 0)
    assert _map_y(y=132, xs=264, binning=1, num_rows=264) == (131, 264)
    assert _map_y(y=263, xs=264, binning=1, num_rows=264) == (0, 264)


@pytest.mark.with_numba
@pytest.mark.parametrize(
    'binning', [1, 2, 4],
)
def test_decode(binning):
    out = np.zeros((8, 8, 264), dtype=np.uint16)
    reads = [
        np.random.randint(low=1, high=1024, size=(1, 264), dtype=np.uint16)
        for i in range(out.shape[0] * out.shape[1] // binning)
    ]

    decoder = FRMS6Decoder(binning=binning)
    decode = decoder.get_decode(native_dtype="u2", read_dtype=np.float32)

    for idx, read in enumerate(reads):
        decode(
            inp=read,
            out=out,
            idx=idx,
            native_dtype=np.uint16,
            rr=None,
            origin=np.array([0, 0, 0]),
            shape=np.array(out.shape),
            ds_shape=np.array([1024, 264, 264]),
        )

    for idx, px in enumerate(out.reshape((-1,))):
        assert not np.isclose(px, 0)


@needsdata
@pytest.mark.with_numba
def test_with_roi(default_frms6, lt_ctx):
    udf = PickUDF()
    roi = np.zeros(default_frms6.shape.nav, dtype=bool)
    roi[0, 0] = 1
    res = lt_ctx.run_udf(udf=udf, dataset=default_frms6, roi=roi)
    assert np.array(res['intensity']).shape == (1, 264, 264)


@needsdata
def test_read_invalid_tileshape(default_frms6):
    partitions = default_frms6.get_partitions()
    p = next(partitions)

    tileshape = Shape(
        (1, 3, 264),
        sig_dims=2,
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=default_frms6.shape,
    )

    tiles = p.get_tiles(tiling_scheme=tiling_scheme)
    with pytest.raises(ValueError):
        next(tiles)

    for _ in tiles:
        pass


@needsdata
def test_scheme_too_large(default_frms6):
    partitions = default_frms6.get_partitions()
    p = next(partitions)
    depth = p.shape[0]

    # we make a tileshape that is too large for the partition here:
    tileshape = Shape(
        (depth + 1,) + tuple(default_frms6.shape.sig),
        sig_dims=default_frms6.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=default_frms6.shape,
    )

    # tile shape is clamped to partition shape:
    tiles = p.get_tiles(tiling_scheme=tiling_scheme)
    t = next(tiles)
    assert tuple(t.tile_slice.shape) == tuple((depth,) + default_frms6.shape.sig)

    for _ in tiles:
        pass


@needsdata
def test_positive_sync_offset(default_frms6, lt_ctx):
    udf = PickUDF()
    sync_offset = 2

    roi = np.zeros(default_frms6.shape.nav, dtype=bool)
    flat_roi = reshaped_view(roi, -1)
    flat_roi[2:10] = True

    ref = lt_ctx.run_udf(dataset=default_frms6, udf=udf, roi=roi)

    ds = lt_ctx.load(
        "frms6", path=FRMS6_TESTDATA_PATH, nav_shape=(4, 2), sync_offset=sync_offset,
        enable_offset_correction=True
    )
    result = lt_ctx.run_udf(dataset=ds, udf=udf)

    assert np.allclose(
        result['intensity'].raw_data,
        ref['intensity'].raw_data
    )


@needsdata
def test_negative_sync_offset(default_frms6, lt_ctx):
    udf = PickUDF()
    sync_offset = -2

    roi = np.zeros(default_frms6.shape.nav, dtype=bool)
    flat_roi = reshaped_view(roi, -1)
    flat_roi[:6] = True

    ref = lt_ctx.run_udf(dataset=default_frms6, udf=udf, roi=roi)

    ds = lt_ctx.load(
        "frms6", path=FRMS6_TESTDATA_PATH, nav_shape=(4, 2), sync_offset=sync_offset,
        enable_offset_correction=True
    )
    result = lt_ctx.run_udf(dataset=ds, udf=udf)

    assert np.allclose(
        result['intensity'].raw_data[abs(sync_offset):],
        ref['intensity'].raw_data
    )


@needsdata
def test_offset_smaller_than_image_count(lt_ctx):
    sync_offset = -65540

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "frms6",
            path=FRMS6_TESTDATA_PATH,
            sync_offset=sync_offset
        )
    assert e.match(
        r"offset should be in \(-65536, 65536\), which is \(-image_count, image_count\)"
    )


@needsdata
def test_offset_greater_than_image_count(lt_ctx):
    sync_offset = 65540

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "frms6",
            path=FRMS6_TESTDATA_PATH,
            sync_offset=sync_offset
        )
    assert e.match(
        r"offset should be in \(-65536, 65536\), which is \(-image_count, image_count\)"
    )


@needsdata
def test_reshape_nav(default_frms6, lt_ctx):
    udf = PickUDF()

    roi = np.zeros(default_frms6.shape.nav, dtype=bool)
    flat_roi = reshaped_view(roi, -1)
    flat_roi[:8] = True

    ref = lt_ctx.run_udf(dataset=default_frms6, udf=udf, roi=roi)

    ds_1 = lt_ctx.load(
        "frms6",
        path=FRMS6_TESTDATA_PATH,
        nav_shape=(8,),
        enable_offset_correction=True
    )
    result_1 = lt_ctx.run_udf(dataset=ds_1, udf=udf)
    shape_1 = lt_ctx.run_udf(dataset=ds_1, udf=SumSigUDF())

    assert shape_1['intensity'].data.shape == (8, )
    assert np.allclose(result_1['intensity'].raw_data, ref['intensity'].raw_data)

    ds_2 = lt_ctx.load(
        "frms6",
        path=FRMS6_TESTDATA_PATH,
        nav_shape=(2, 2, 2),
        enable_offset_correction=True
    )
    result_2 = lt_ctx.run_udf(dataset=ds_2, udf=udf)
    shape_2 = lt_ctx.run_udf(dataset=ds_2, udf=SumSigUDF())
    assert shape_2['intensity'].data.shape == (2, 2, 2)
    assert np.allclose(result_2['intensity'].raw_data, ref['intensity'].raw_data)


@needsdata
def test_incorrect_sig_shape(lt_ctx):
    sig_shape = (5, 5)

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "frms6",
            path=FRMS6_TESTDATA_PATH,
            sig_shape=sig_shape
        )
    assert e.match(
        r"sig_shape must be of size: 69696"
    )


@needsdata
def test_compare_backends(lt_ctx, default_frms6, buffered_frms6):
    y = random.choice(range(default_frms6.shape.nav[0]))
    x = random.choice(range(default_frms6.shape.nav[1]))
    mm_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=default_frms6,
        x=x, y=y,
    )).intensity
    buffered_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=buffered_frms6,
        x=x, y=y,
    )).intensity

    assert np.allclose(mm_f0, buffered_f0)


@needsdata
@pytest.mark.skipif(
    sys.platform.startswith("darwin"),
    reason="No support for direct I/O on Mac OS X"
)
def test_compare_direct_to_mmap(lt_ctx, default_frms6, direct_frms6):
    y = random.choice(range(default_frms6.shape.nav[0]))
    x = random.choice(range(default_frms6.shape.nav[1]))
    mm_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=default_frms6,
        x=x, y=y,
    )).intensity
    buffered_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=direct_frms6,
        x=x, y=y,
    )).intensity

    assert np.allclose(mm_f0, buffered_f0)


@needsdata
@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_compare_backends_sparse(lt_ctx, default_frms6, buffered_frms6, as_sparse):
    roi = np.zeros(default_frms6.shape.nav, dtype=bool).reshape((-1,))
    roi[0] = True
    roi[1] = True
    roi[16] = True
    roi[32] = True
    roi[-1] = True
    if as_sparse:
        roi = roi_as_sparse(roi)
    mm_f0 = lt_ctx.run_udf(dataset=default_frms6, udf=PickUDF(), roi=roi)['intensity']
    buffered_f0 = lt_ctx.run_udf(dataset=buffered_frms6, udf=PickUDF(), roi=roi)['intensity']

    assert np.allclose(mm_f0, buffered_f0)


def test_bad_params(ds_params_tester, standard_bad_ds_params):
    args = ("frms6", FRMS6_TESTDATA_PATH)
    for params in standard_bad_ds_params:
        ds_params_tester(*args, **params)



from empyre.vis.colors import ColormapCubehelix, ColormapPerception, ColormapHLS, ColormapClassic

from .base import visualize_simple, get_plottable_2D_channels

from libertem.common.viz import encode_image


__all__ = ['ColormapCubehelix', 'ColormapPerception', 'ColormapHLS',
           'ColormapClassic', 'cmaps', 'CMAP_CIRCULAR_DEFAULT',
           'visualize_simple', 'encode_image', 'get_plottable_2D_channels']


cmaps = {'cubehelix_standard': ColormapCubehelix(),
         'cubehelix_reverse': ColormapCubehelix(reverse=True),
         'cubehelix_circular': ColormapCubehelix(start=1, rot=1,
                                                 minLight=0.5, maxLight=0.5, sat=2),
         'perception_circular': ColormapPerception(),
         'hls_circular': ColormapHLS(),
         'classic_circular': ColormapClassic()}

CMAP_CIRCULAR_DEFAULT = cmaps['cubehelix_circular']



import os
import contextlib
from copy import deepcopy
import functools
import logging
import signal
from typing import Any, Optional, Union, Callable
from collections.abc import Iterable
import uuid

from dask import distributed as dd
import dask

from libertem.common.threading import set_num_threads_env

from .base import BaseJobExecutor, AsyncAdapter, ResourceError
from libertem.common.executor import (
    JobCancelledError, TaskCommHandler, TaskProtocol, Environment, WorkerContext,
)
from libertem.common.async_utils import sync_to_async
from libertem.common.scheduler import Worker, WorkerSet
from libertem.common.backend import set_use_cpu, set_use_cuda
from libertem.common.async_utils import adjust_event_loop_policy
from .utils import assign_cudas

log = logging.getLogger(__name__)


class DaskWorkerContext(WorkerContext):
    def __init__(self, comms_topic: Optional[str]):
        # DaskWorkerContext sends all messages via a single unique topic id
        # which are later unpacked on the client node; this allows us to handle
        # concurrent runs using the same executor with separate comms channels
        self._comms_topic = comms_topic

    @property
    def dask_worker(self):
        try:
            return self._worker
        except AttributeError:
            self._worker = dd.get_worker()
            return self._worker

    def signal(self, ident: str, topic: str, msg_dict: dict[str, Any]):
        if self._comms_topic is None:
            # Scheduler Dask does not have comms so don't send
            return
        msg_dict.update({'ident': ident, 'topic': topic})
        try:
            self.dask_worker.log_event(self._comms_topic, msg_dict)
        except AttributeError:
            # No structured logs available in this Dask
            # Catch the exception here just in case there is
            # version / API mismatch
            pass


@contextlib.contextmanager
def set_worker_log_level(level: Union[str, int], force: bool = False):
    """
    Set the dask.distributed log level for any processes spawned
    within the context manager. If force is False, don't overwrite
    any existing environment variable.
    """
    env_keys = ['DASK_DISTRIBUTED__LOGGING__DISTRIBUTED']
    try:
        old_env = {
            k: os.environ[k]
            for k in env_keys
            if k in os.environ
        }
        os.environ.update({
            k: str(level)
            for k in env_keys
            if force or (k not in old_env)
        })
        yield
    finally:
        os.environ.update(old_env)
        for key in env_keys:
            if key not in old_env:
                del os.environ[key]


def worker_setup(resource, device):
    # Disable handling Ctrl-C on the workers for a local cluster
    # since the nanny restarts workers in that case and that gets mixed
    # with Ctrl-C handling of the main process, at least on Windows
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    if resource == "CUDA":
        set_use_cuda(device)
    elif resource == "CPU":
        set_use_cpu(device)
    else:
        raise ValueError("Unknown resource %s, use 'CUDA' or 'CPU'", resource)


def cluster_spec(
    cpus: Union[int, Iterable[int]],
    cudas: Union[int, Iterable[int]],
    has_cupy: bool,
    name: str = 'default',
    num_service: int = 1,
    options: Optional[dict] = None,
    preload: Optional[tuple[str, ...]] = None
):
    '''
    Create a worker specification dictionary for a LiberTEM Dask cluster

    The return from this function can be passed to :code:`DaskJobExecutor.make_local(spec=spec)`.

    This creates a Dask cluster spec with special initializations and resource tags
    for CPU + GPU processing in LiberTEM.
    See :ref:`cluster spec` for an example.
    See https://distributed.dask.org/en/stable/api.html#distributed.SpecCluster
    for more info on cluster specs.

    Parameters
    ----------
    cpus: int | Iterable[int]
        IDs for CPU workers as an iterable, or an integer number of workers to create.
        Currently no pinning is used, i.e. this specifies the total
        number and identification of workers, not the CPU cores that are used.
    cudas: int | Iterable[int]
        IDs for CUDA device workers as an iterable, or an integer number of GPU workers to
        create. LiberTEM will use the IDs specified or assign round-robin to the available devices.
        In the iterable case these have to match CUDA device IDs on the system.
        Specify the same ID multiple times to spawn multiple workers on the same CUDA device.
    has_cupy: bool
        Specify if the cluster should signal that it supports GPU-based array programming using
        CuPy
    name
        Prefix for the worker names
    num_service
        Number of additional workers that are reserved for service tasks. Computation tasks
        will not be scheduled on these workers, which guarantees responsive behavior for file
        browsing etc.
    options
        Options to pass through to every worker. See Dask documentation for details
    preload
        Items to preload on workers in addition to LiberTEM-internal preloads.
        This can be used to load libraries, for example HDF5 filter plugins before h5py is used.
        See https://docs.dask.org/en/stable/how-to/customize-initialization.html#preload-scripts
        for more information.

    See also
    --------
    :func:`libertem.utils.devices.detect`
    '''
    if options is None:
        options = {}
    if preload is None:
        preload = ()
    if options.get("nthreads") is None:
        options["nthreads"] = 1
    if options.get("silence_logs") is None:
        options["silence_logs"] = logging.WARN

    workers_spec = {}

    cpu_options = deepcopy(options)
    cpu_options["resources"] = {"CPU": 1, "compute": 1, "ndarray": 1}
    cpu_base_spec = {
        "cls": dd.Nanny,
        "options": cpu_options,
    }

    # Service workers not for computation

    service_options = deepcopy(options)
    service_options["resources"] = {}
    service_base_spec = {
        "cls": dd.Nanny,
        "options": service_options
    }

    cuda_options = deepcopy(options)
    cuda_options["resources"] = {"CUDA": 1, "compute": 1}
    if has_cupy:
        cuda_options["resources"]["ndarray"] = 1
    cuda_base_spec = {
        "cls": dd.Nanny,
        "options": cuda_options
    }

    def _get_tracing_setup(service_name: str, service_id: str) -> str:
        return (
            f"from libertem.common.tracing import maybe_setup_tracing; "
            f"maybe_setup_tracing(service_name='{service_name}', service_id='{service_id}')"
        )

    if isinstance(cpus, int):
        cpus = tuple(range(cpus))

    for cpu in cpus:
        worker_name = f'{name}-cpu-{cpu}'
        cpu_spec = deepcopy(cpu_base_spec)
        cpu_spec['options']['preload'] = preload + (
            'from libertem.executor.dask import worker_setup; '
            + f'worker_setup(resource="CPU", device={cpu})',
            _get_tracing_setup(worker_name, str(cpu)),
            'libertem.preload',
        )
        workers_spec[worker_name] = cpu_spec

    for service in range(num_service):
        worker_name = f'{name}-service-{service}'
        service_spec = deepcopy(service_base_spec)
        service_spec['options']['preload'] = preload + (
            _get_tracing_setup(worker_name, str(service)),
            'libertem.preload',
        )
        workers_spec[worker_name] = service_spec

    cudas = assign_cudas(cudas)

    for cuda in cudas:
        worker_name = f'{name}-cuda-{cuda}'
        if worker_name in workers_spec:
            num_with_name = sum(n.startswith(worker_name) for n in workers_spec)
            worker_name = f'{worker_name}-{num_with_name - 1}'
        cuda_spec = deepcopy(cuda_base_spec)
        cuda_spec['options']['preload'] = preload + (
            'from libertem.executor.dask import worker_setup; '
            + f'worker_setup(resource="CUDA", device={cuda})',
            _get_tracing_setup(worker_name, str(cuda)),
            'libertem.preload',
        )
        workers_spec[worker_name] = cuda_spec

    return workers_spec


def _run_task(task, params, task_id, threaded_executor, comms_topic: Optional[str]):
    """
    Very simple wrapper function. As dask internally caches functions that are
    submitted to the cluster in various ways, we need to make sure to
    consistently use the same function, and not build one on the fly.

    Without this function, UDFTask->UDF->UDFData ends up in the
    cache, which blows up memory usage over time.
    """
    worker_context = DaskWorkerContext(comms_topic)
    env = Environment(threads_per_worker=1,
                      threaded_executor=threaded_executor,
                      worker_context=worker_context)
    task_result = task(env=env, params=params)
    return {
        "task_result": task_result,
        "task_id": task_id,
    }


def _simple_run_task(task):
    return task()


class CommonDaskMixin:
    client: dd.Client

    def _task_idx_to_workers(self, workers: WorkerSet, idx: int) -> WorkerSet:
        hosts = list(sorted(workers.hosts()))
        host_idx = idx % len(hosts)
        host = hosts[host_idx]
        return workers.filter(lambda w: w.host == host)

    def _future_for_location(
        self,
        task: TaskProtocol,
        locations: WorkerSet,
        resources,
        workers: WorkerSet,
        task_args=None,
        wrap_fn=_run_task,
    ):
        """
        Submit tasks and return the resulting futures

        Parameters
        ----------
        task:
            callable
        locations:
            potential locations to run the task
        resources:
            required resources for the task
        workers : WorkerSet
            Available workers in the cluster
        """
        submit_kwargs = {}
        if task_args is None:
            task_args = {}
        if locations is not None:
            if len(locations) == 0:
                raise ValueError("no workers found for task")
            location_names = locations.names()
        else:
            location_names = None
        submit_kwargs.update({
            'resources': self._validate_resources(workers, resources),
            'workers': location_names,
            'pure': False,
        })
        return self.client.submit(
            wrap_fn, task, *task_args, **submit_kwargs
        )

    def _validate_resources(self, workers, resources):
        # This is set in the constructor of DaskJobExecutor
        if self.lt_resources:
            if not self._resources_available(workers, resources):
                raise ResourceError("Requested resources not available in cluster:", resources)
            result = resources
        else:
            if 'CUDA' in resources:
                raise ResourceError(
                    "Requesting CUDA resource on a cluster without resource management."
                )
            result = {}
        return result

    def _resources_available(self, workers, resources):
        def filter_fn(worker):
            return all(worker.resources.get(key, 0) >= resources[key] for key in resources.keys())

        return len(workers.filter(filter_fn))

    def has_libertem_resources(self):
        workers = self.get_available_workers()

        def has_resources(worker):
            r = worker.resources
            return 'compute' in r and (('CPU' in r and 'ndarray' in r) or 'CUDA' in r)

        return len(workers.filter(has_resources)) > 0

    def _get_future(
        self,
        task: TaskProtocol,
        workers: WorkerSet,
        idx: int,
        params_handle,
        threaded_executor,
        comms_topic: Optional[str]
    ):
        if len(workers) == 0:
            raise RuntimeError("no workers available!")
        params_fut = self._scatter_map[params_handle]
        return self._future_for_location(
            task=task,
            locations=task.get_locations() or self._task_idx_to_workers(
                workers, idx
            ),
            resources=task.get_resources(),
            workers=workers,
            task_args=(
                params_fut,
                idx,
                threaded_executor,
                comms_topic,
            )
        )

    def get_available_workers(self) -> WorkerSet:
        info = self.client.scheduler_info()
        return WorkerSet([
            Worker(
                name=worker['name'],
                host=worker['host'],
                resources=worker['resources'],
                nthreads=worker['nthreads'],
            )
            for worker in info['workers'].values()
        ])

    def get_resource_details(self):
        workers = self.get_available_workers()
        details = {}

        for worker in workers:
            host_name = worker.host
            if worker.name.startswith("tcp"):
                # for handling `dask-worker`
                # `dask-worker` name starts with "tcp"
                #  only supports CPU
                resource = 'cpu'
            else:
                # for handling `libertem-worker`
                r = worker.resources
                if "CPU" in r:
                    resource = 'cpu'
                elif "CUDA" in r:
                    resource = 'cuda'
                else:
                    resource = 'service'

            if host_name not in details.keys():
                details[host_name] = {
                                 'host': host_name,
                                 'cpu': 0,
                                 'cuda': 0,
                                 'service': 0,
                            }
            details[host_name][resource] += 1

        details_sorted = []
        for host in sorted(details.keys()):
            details_sorted.append(details[host])

        return details_sorted


def _dispatch_messages(subscribers: dict[str, list[Callable]], dask_message: tuple[float, dict]):
    """
    Unpacks the Dask message format and forwards the message
    to all subscribed callbacks for that topic (if any)
    """
    timestamp, message = dask_message
    true_topic = message.pop('topic')
    for handler in subscribers.get(true_topic, []):
        handler(true_topic, message)


class DaskJobExecutor(CommonDaskMixin, BaseJobExecutor):
    '''
    Default LiberTEM executor that uses `Dask futures
    <https://docs.dask.org/en/stable/futures.html>`_.

    Parameters
    ----------

    client : distributed.Client
    is_local : bool
        Close the Client and cluster when the executor is closed.
    lt_resources : bool
        Specify if the cluster has LiberTEM resource tags and environment
        variables for GPU processing. Autodetected by default.
    '''
    def __init__(self, client: dd.Client, is_local: bool = False,
                lt_resources: bool = None):
        self.is_local = is_local
        self.client = client
        if lt_resources is None:
            lt_resources = self.has_libertem_resources()
        self.lt_resources = lt_resources
        self._futures = {}
        self._scatter_map = {}

    @contextlib.contextmanager
    def scatter(self, obj):
        # an additional layer of indirection, because we want to be able to
        # redirect keys to new values
        handle = str(uuid.uuid4())
        try:
            fut = self.client.scatter(obj, broadcast=True, hash=False)
            self._scatter_map[handle] = fut
            yield handle
        finally:
            if handle in self._scatter_map:
                del self._scatter_map[handle]

    def scatter_update(self, handle, obj):
        fut = self.client.scatter(obj, broadcast=True, hash=False)
        self._scatter_map[handle] = fut

    def scatter_update_patch(self, handle, patch):
        fut = self._scatter_map[handle]

        def _do_patch(obj):
            if not hasattr(obj, 'patch'):
                raise TypeError(f'object is not patcheable: {obj}')
            obj.patch(patch)

        # can't `client.run` here, as that doesn't resolve the scatter future
        futures = []
        for worker in self.get_available_workers().names():
            futures.append(self.client.submit(_do_patch, fut, pure=False, workers=[worker]))
        for res in dd.as_completed(futures, loop=self.client.loop):
            pass

    def run_tasks(
        self,
        tasks: Iterable[TaskProtocol],
        params_handle: Any,
        cancel_id: Any,
        task_comm_handler: TaskCommHandler,
    ):
        tasks = list(tasks)
        tasks_w_index = list(enumerate(tasks))

        def _id_to_task(task_id):
            return tasks[task_id]

        workers = self.get_available_workers()
        threaded_executor = workers.has_threaded_workers()

        self._futures[cancel_id] = []
        initial = []

        try:
            topic_id = f'topic-{cancel_id}'
            # Wrap all subscriptions into single unique topic
            self.client.subscribe_topic(
                topic_id,
                functools.partial(_dispatch_messages,
                                  task_comm_handler.subscriptions)
            )
        except AttributeError:
            # Dask version does not support structured logs
            # Fall back to partition-level progress updates only
            topic_id = None

        for w in range(int(len(workers))):
            if not tasks_w_index:
                break
            idx, wrapped_task = tasks_w_index.pop(0)
            future = self._get_future(wrapped_task, workers, idx, params_handle,
                                      threaded_executor, topic_id)
            initial.append(future)
            self._futures[cancel_id].append(future)

        try:
            as_completed = dd.as_completed(initial, with_results=True, loop=self.client.loop)
            for future, result_wrap in as_completed:
                if future.cancelled():
                    log.debug(
                        "future %r is cancelled, stopping",
                        future,
                    )
                    del self._futures[cancel_id]
                    raise JobCancelledError()
                result = result_wrap['task_result']
                task = _id_to_task(result_wrap['task_id'])
                if tasks_w_index:
                    idx, wrapped_task = tasks_w_index.pop(0)
                    future = self._get_future(
                        wrapped_task, workers, idx, params_handle, threaded_executor, topic_id
                    )
                    as_completed.add(future)
                    self._futures[cancel_id].append(future)
                yield result, task
        finally:
            if cancel_id in self._futures:
                del self._futures[cancel_id]
            if topic_id is not None:
                self.client.unsubscribe_topic(topic_id)

    def cancel(self, cancel_id):
        log.debug("cancelling with cancel_id=`%s`", cancel_id)
        if cancel_id in self._futures:
            futures = self._futures[cancel_id]
            self.client.cancel(futures)

    def run_each_partition(self, partitions, fn, all_nodes=False):
        """
        Run `fn` for all partitions. Yields results in order of completion.

        Parameters
        ----------

        partitions : List[Partition]
            List of relevant partitions.

        fn : callable
            Function to call, will get the partition as first and only argument.

        all_nodes : bool
            If all_nodes is True, run the function on all nodes that have this partition,
            otherwise run on any node that has the partition. If a partition has no location,
            the function will not be run for that partition if `all_nodes` is True, otherwise
            it will be run on any node.
        """
        def _make_items_all():
            for p in partitions:
                locs = p.get_locations()
                if locs is None:
                    continue
                for workers in locs.group_by_host():
                    yield (lambda: fn(p), workers)

        if all_nodes:
            items = _make_items_all()
        else:
            # TODO check if we should request a compute resource
            items = ((lambda: fn(p), p.get_locations(), {})
                     for p in partitions)
        workers = self.get_available_workers()
        futures = [
            self._future_for_location(*item, workers, wrap_fn=_simple_run_task)
            for item in items
        ]
        # TODO: do we need cancellation and all that good stuff?
        for future, result in dd.as_completed(futures, with_results=True, loop=self.client.loop):
            if future.cancelled():
                raise JobCancelledError()
            yield result

    def run_function(self, fn, *args, **kwargs):
        """
        run a callable :code:`fn` on any worker
        """
        fn_with_args = functools.partial(fn, *args, **kwargs)
        future = self.client.submit(fn_with_args, priority=1, pure=False)
        return future.result()

    def map(self, fn, iterable):
        """
        Run a callable :code:`fn` for each element in :code:`iterable`, on arbitrary worker nodes.

        Parameters
        ----------

        fn : callable
            Function to call. Should accept exactly one parameter.

        iterable : Iterable
            Which elements to call the function on.
        """
        return [future.result()
                for future in self.client.map(fn, iterable, pure=False)]

    def run_each_host(self, fn, *args, **kwargs):
        """
        Run a callable :code:`fn` once on each host, gathering all results into
        a dict host -> result
        """
        # TODO: any cancellation/errors to handle?
        available_workers = self.get_available_workers()

        future_map = {}
        for worker_set in available_workers.group_by_host():
            future_map[worker_set.example().host] = self.client.submit(
                functools.partial(fn, *args, **kwargs),
                priority=1,
                workers=worker_set.names(),
                # NOTE: need pure=False, otherwise the functions will all map to the same
                # scheduler key and will only run once
                pure=False,
            )
        result_map = {
            host: future.result()
            for host, future in future_map.items()
        }
        return result_map

    def run_each_worker(self, fn, *args, **kwargs):
        # Client.run() creates issues on Windows and OS X with Python 3.6
        # FIXME workaround may not be needed anymore for Python 3.7+
        available_workers = self.get_available_workers()

        future_map = {}
        for worker in available_workers:
            future_map[worker.name] = self.client.submit(
                functools.partial(fn, *args, **kwargs),
                priority=1,
                workers=[worker.name],
                # NOTE: need pure=False, otherwise the functions will all map to the same
                # scheduler key and will only run once
                pure=False,
            )
        result_map = {
            name: future.result()
            for name, future in future_map.items()
        }
        return result_map

    def close(self):
        if self.is_local:
            # Client.close won't close the Cluster itself because
            # we provided an external dd.SpecCluster
            self.client.close()
            # Manually close the cluster if not yet torn down
            # use getattr just in case cluster is already gone
            if getattr(self.client, 'cluster', None) is not None:
                self.client.cluster.close(timeout=30)
        # NOTE: distributed already registers atexit handlers for
        # both clients and clusters, this is here to allow manual closure
        # followed by creation of a new Executor without accumulating clusters

    @classmethod
    def connect(cls, scheduler_uri, *args, client_kwargs: Optional[dict] = None, **kwargs):
        """
        Connect to a remote dask scheduler.

        Parameters
        ----------
        scheduler_uri: str
            Compatible with the :code:`address` parameter of :class:`distributed.Client`.
        client_kwargs: dict or None
            Passed as kwargs to :class:`distributed.Client`.
            :code:`client_kwargs['set_as_default']` is set to :code:`False`
            unless specified otherwise to avoid interference with Dask-based workflows.
            Pass :code:`client_kwargs={'set_as_default': True}` to set the Client as the
            default Dask scheduler and keep it running when the Context closes.
        *args, **kwargs: Passed to :class:`DaskJobExecutor`.

        Returns
        -------
        DaskJobExecutor
            the connected JobExecutor
        """
        if client_kwargs is None:
            client_kwargs = {}
        if client_kwargs.get('set_as_default') is None:
            client_kwargs['set_as_default'] = False
        is_local = not client_kwargs['set_as_default']
        client = dd.Client(address=scheduler_uri, **client_kwargs)
        return cls(client=client, is_local=is_local, *args, **kwargs)

    @classmethod
    def make_local(cls, spec: Optional[dict] = None, cluster_kwargs: Optional[dict] = None,
            client_kwargs: Optional[dict] = None, preload: Optional[tuple[str]] = None):
        """
        Spin up a local dask cluster

        Parameters
        ----------
        spec
            Dask cluster spec, see
            https://distributed.dask.org/en/stable/api.html#distributed.SpecCluster
            for more info.
            :func:`libertem.utils.devices.detect` allows to detect devices that can be used
            with LiberTEM, and :func:`cluster_spec` can be used to create a :code:`spec`
            with customized parameters.
        cluster_kwargs
            Passed to :class:`distributed.SpecCluster`.
        client_kwargs
            Passed to :class:`distributed.Client`. Pass
            :code:`client_kwargs={'set_as_default': True}` to set the Client as the
            default Dask scheduler.
        preload: Optional[Tuple[str]]
            Passed to :func:`cluster_spec` if :code:`spec` is :code:`None`.

        Returns
        -------
        DaskJobExecutor
            the connected JobExecutor
        """

        # Distributed doesn't adjust the event loop policy when being run
        # from within pytest as of version 2.21.0. For that reason we
        # adjust the policy ourselves here.
        adjust_event_loop_policy()

        if spec is None:
            from libertem.utils.devices import detect
            spec = cluster_spec(**detect(), preload=preload)
        else:
            if preload is not None:
                raise ValueError(
                    "Passing both spec and preload is not supported. "
                    "Instead, include preloading specification in the spec"
                )
        if client_kwargs is None:
            client_kwargs = {}
        if client_kwargs.get('set_as_default') is None:
            client_kwargs['set_as_default'] = False

        if cluster_kwargs is None:
            cluster_kwargs = {}
        if cluster_kwargs.get('silence_logs') is None:
            cluster_kwargs['silence_logs'] = logging.WARN

        dist_log_level = dask.config.get('distributed.logging.distributed', default=None)
        if dist_log_level is None:
            dist_log_level = cluster_kwargs['silence_logs']

        with set_num_threads_env(n=1), set_worker_log_level(dist_log_level):
            # Mitigation for https://github.com/dask/distributed/issues/6776
            with dask.config.set({"distributed.worker.profile.enabled": False}):
                cluster = dd.SpecCluster(workers=spec, **(cluster_kwargs or {}))
                client = dd.Client(cluster, **(client_kwargs or {}))
                client.wait_for_workers(len(spec))

        is_local = not client_kwargs['set_as_default']

        return cls(client=client, is_local=is_local, lt_resources=True)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class AsyncDaskJobExecutor(AsyncAdapter):
    def __init__(self, wrapped=None, *args, **kwargs):
        if wrapped is None:
            wrapped = DaskJobExecutor(*args, **kwargs)
        super().__init__(wrapped)

    @classmethod
    async def connect(cls, scheduler_uri, *args, **kwargs):
        executor = await sync_to_async(functools.partial(
            DaskJobExecutor.connect,
            scheduler_uri=scheduler_uri,
            *args,
            **kwargs,
        ))
        return cls(wrapped=executor)

    @classmethod
    async def make_local(cls, spec=None, cluster_kwargs=None, client_kwargs=None):
        executor = await sync_to_async(functools.partial(
            DaskJobExecutor.make_local,
            spec=spec,
            cluster_kwargs=cluster_kwargs,
            client_kwargs=client_kwargs,
        ))
        return cls(wrapped=executor)


def cli_worker(
    scheduler,
    local_directory,
    cpus,
    cudas,
    has_cupy,
    name,
    log_level,
    preload: tuple[str, ...]
):
    import asyncio

    options = {
        "silence_logs": log_level,
        "local_directory": local_directory
    }

    spec = cluster_spec(
        cpus=cpus, cudas=cudas, has_cupy=has_cupy, name=name,
        options=options, preload=preload,
    )

    async def run(spec):
        # Mitigation for https://github.com/dask/distributed/issues/6776
        with dask.config.set({"distributed.worker.profile.enabled": False}):
            workers = []
            for name, spec in spec.items():
                cls = spec['cls']
                workers.append(
                    cls(scheduler, name=name, **spec['options'])
                )
            import asyncio
            await asyncio.gather(*workers)
            for w in workers:
                await w.finished()

    asyncio.get_event_loop().run_until_complete(run(spec))



import sys
import time
from itertools import chain
import gc

import numpy as np
import pytest

from libertem.udf.masks import ApplyMasksUDF


# based on https://code.activestate.com/recipes/577504/
def total_size(o):
    def dict_handler(d):
        return chain.from_iterable(d.items())

    def object_handler(o):
        if hasattr(o, '__dict__'):
            return dict_handler(o.__dict__)
        else:
            return tuple()

    all_handlers = {
        tuple: iter,
        list: iter,
        # deque: iter, Triggers RuntimeError with Dask client
        dict: dict_handler,
        set: iter,
        frozenset: iter,
        object: object_handler,
    }
    seen = set()                      # track which object id's have already been seen
    default_size = sys.getsizeof(0)       # estimate sizeof object without __sizeof__

    def sizeof(o):
        if id(o) in seen:       # do not double count the same object
            return 0
        seen.add(id(o))
        s = sys.getsizeof(o, default_size)

        for typ, handler in all_handlers.items():
            if isinstance(o, typ):
                s += sum(map(sizeof, handler(o)))
                break
        return s

    return sizeof(o)


def worker_memory(client):
    info = client.scheduler_info()
    memory = 0
    for name, w_info in info['workers'].items():
        memory += w_info['metrics']['memory']
    return memory


@pytest.mark.slow
@pytest.mark.parametrize(
    'ctx_select', ('dask', 'inline')
)
def test_executor_memleak(local_cluster_ctx, lt_ctx_fast, default_raw, ctx_select):
    if ctx_select == 'dask':
        ctx = local_cluster_ctx
        rounds = 5

        def get_worker_mem(ctx):
            return worker_memory(ctx.executor.client)

    elif ctx_select == 'inline':
        ctx = lt_ctx_fast
        rounds = 1

        def get_worker_mem(ctx):
            return 0

    mask_count = 8*1014*1024 // np.prod(default_raw.shape.sig)

    mask_shape = (mask_count, *default_raw.shape.sig)
    masks = np.zeros(mask_shape)

    # Intentionally "bad" factory function: make it large by closing over masks
    def mask_factory():
        return masks

    udf = ApplyMasksUDF(
        mask_factories=mask_factory,
        mask_count=mask_count,
        mask_dtype=masks.dtype,
        use_torch=False
    )

    # warm-up
    for _ in range(2):
        for _ in ctx.run_udf_iter(dataset=default_raw, udf=udf):
            pass

    cumulative_worker_delta = 0
    cumulative_executor_delta = 0

    for round in range(rounds):
        gc.collect()

        # Allow to settle
        time.sleep(1)

        ctx.executor.run_each_worker(gc.collect)

        executor_size_before = total_size(ctx)
        worker_mem_before = get_worker_mem(ctx)

        executor_size_during = None

        for res in ctx.run_udf_iter(dataset=default_raw, udf=udf):
            if executor_size_during is None:
                executor_size_during = total_size(ctx)
                worker_mem_during = get_worker_mem(ctx)

        gc.collect()

        # Allow to settle
        time.sleep(1)

        ctx.executor.run_each_worker(gc.collect)

        # Allow to settle
        time.sleep(1)

        executor_size_after = total_size(ctx)
        worker_mem_after = get_worker_mem(ctx)

        active_use = worker_mem_during - worker_mem_before

        # Memory use does increase slowly. Just make sure it is not caused by keeping
        # a big array around
        worker_delta = worker_mem_after - worker_mem_before
        executor_delta = executor_size_after - executor_size_before

        print(f"Round {round}")
        print(f"Memory use during UDF run: {active_use}.")
        print(f"Memory increase worker: {worker_delta}.")
        print(f"Memory increase executor: {executor_delta}.")

        cumulative_worker_delta += worker_delta
        cumulative_executor_delta += executor_delta

    worker_count = len(ctx.executor.get_available_workers())

    assert cumulative_worker_delta/rounds/worker_count < sys.getsizeof(masks) * 0.1
    assert cumulative_executor_delta/rounds < sys.getsizeof(masks) * 0.1



import logging

import numpy as np

from .base import Live2DPlot


logger = logging.getLogger(__name__)


class BQLive2DPlot(Live2DPlot):
    """
    bqplot-image-gl-based live plot.

    .. versionadded:: 0.7.0

    Parameters
    ----------
    dataset : DataSet
        The dataset on which the UDF will be run. This allows
        to determine the shape of the plots for initialization.

    udf : UDF
        The UDF instance this plot is associated to. This needs to be
        the same instance that is passed to
        :meth:`~libertem.api.Context.run_udf`.

    roi : numpy.ndarray or None
        Region of interest (ROI) that the UDF will
        be run on. This is necessary for UDFs where the `extra_shape`
        parameter of result buffers is a function of the ROI, such as
        :class:`~libertem.udf.raw.PickUDF`.

    channel : misc
        Indicate the channel to be plotted.

        - :code:`None`: The first plottable (2D) channel of the UDF is plotted.
        - :code:`str`: The UDF result buffer name that should be plotted.
        - :code:`tuple(str, function(ndarray) -> ndarray)`: The UDF result buffer name that
          should be plotted together with a function that extracts a plottable result
        - :code:`function(udf_result, damage) -> (ndarray, damage)`: Function that derives a
          plottable 2D ndarray and damage indicator from the full
          UDF results and the processed nav space. See :ref:`plotting` for more details!

    title : str
        The plot title. By default UDF class name and channel name.

    min_delta : float
        Minimum time span in seconds between updates to reduce overheads for slow plotting.

    udfresult : UDFResults, optional
        UDF result to initialize the plot data and determine plot shape. If None (default),
        this is determined using :meth:`~libertem.udf.base.UDFRunner.dry_run` on the dataset,
        UDF and ROI. This parameter allows re-using buffers to avoid unnecessary dry runs.
    """
    def __init__(
            self, dataset, udf, roi=None, channel=None, title=None, min_delta=1/60, udfresult=None
    ):
        super().__init__(
            dataset=dataset,
            udf=udf,
            roi=roi,
            channel=channel,
            title=title,
            min_delta=min_delta,
            udfresult=udfresult,
        )
        # keep bqplot and bqplot_image_gl as optional dependencies
        from bqplot import Figure, LinearScale, Axis, ColorScale
        from bqplot_image_gl import ImageGL

        scale_x = LinearScale(min=0, max=1)
        # Make sure y points down
        # See https://libertem.github.io/LiberTEM/concepts.html#coordinate-system
        scale_y = LinearScale(min=1, max=0)
        axis_x = Axis(scale=scale_x, label='x')
        axis_y = Axis(scale=scale_y, label='y', orientation='vertical')

        s = self.data.shape
        aspect = s[1] / s[0]

        figure = Figure(
            axes=[axis_x, axis_y],
            scale_x=scale_x,
            scale_y=scale_y,
            min_aspect_ratio=aspect,
            max_aspect_ratio=aspect,
            title=self.title
        )

        color_scale = ColorScale(min=0, max=1)

        scales_image = {'x': scale_x,
                        'y': scale_y,
                        'image': color_scale}

        dtype = np.result_type(self.data, np.int8)
        image = ImageGL(image=self.data.astype(dtype), scales=scales_image)
        figure.marks = (image,)
        self.figure = figure
        self.image = image
        self.color_scale = color_scale

    def display(self):
        from IPython.display import display
        display(self.figure)

    def update(self, damage, force=False):
        dtype = np.result_type(self.data, np.int8)
        # Map on dtype that supports subtraction
        valid_data = self.data[damage].astype(dtype)
        valid_data = valid_data[np.isfinite(valid_data)]
        if valid_data.size > 0:
            mmin = valid_data.min()
            mmax = valid_data.max()
        else:
            mmin = 1
            mmax = 1 + 1e-12
        delta = mmax - mmin
        if delta <= 0:
            delta = 1
        # Map on color scale range 0..1
        self.image.image = (self.data.astype(dtype) - mmin) / delta



import numpy as np
import scipy.sparse as sp
import sparse
import pytest
from sparseconverter import CUPY

from libertem.common.container import MaskContainer
from libertem.io.dataset.base import DataTile
from libertem.common import Slice, Shape
from libertem.masks import gradient_x


@pytest.fixture
def masks():
    input_masks = [
        lambda: np.ones((128, 128)),
        lambda: sparse.zeros((128, 128)),
        lambda: np.ones((128, 128)),
        lambda: sp.csr_matrix(((1,), ((64,), (64,))), shape=(128, 128), dtype=np.float32),
        lambda: gradient_x(128, 128, dtype=np.float32),
    ]
    return MaskContainer(mask_factories=input_masks, dtype=np.float32)


def test_mask_caching_1():
    input_masks = [
        lambda: np.ones((128, 128)),
        lambda: np.zeros((128, 128)),
    ]
    mask_container = MaskContainer(mask_factories=input_masks, dtype="float32")

    shape = Shape((16 * 16, 128, 128), sig_dims=2)
    slice_ = Slice(origin=(0, 0, 0), shape=shape)
    mask_container.get(slice_)

    key = (mask_container.dtype, False, True, 'numpy')

    cache_info = mask_container._get_masks_for_slice[key].cache_info()
    assert cache_info.hits == 0
    assert cache_info.misses == 1

    mask_container.get(slice_)

    cache_info = mask_container._get_masks_for_slice[key].cache_info()
    assert cache_info.hits == 1
    assert cache_info.misses == 1

    slice_ = Slice(origin=(1, 0, 0), shape=shape)

    mask_container.get(slice_)

    cache_info = mask_container._get_masks_for_slice[key].cache_info()
    assert cache_info.hits == 2
    assert cache_info.misses == 1


def test_mask_caching_2():
    input_masks = [
        lambda: np.ones((128, 128)),
        lambda: np.zeros((128, 128)),
    ]
    mask_container = MaskContainer(mask_factories=input_masks, dtype="float32")

    shape1 = Shape((16 * 16, 128, 128), sig_dims=2)
    shape2 = Shape((8 * 16, 128, 128), sig_dims=2)
    slice_ = Slice(origin=(0, 0, 0), shape=shape1)
    mask_container.get(slice_)

    key = (mask_container.dtype, False, True, 'numpy')

    cache_info = mask_container._get_masks_for_slice[key].cache_info()
    assert cache_info.hits == 0
    assert cache_info.misses == 1

    mask_container.get(slice_)

    cache_info = mask_container._get_masks_for_slice[key].cache_info()
    assert cache_info.hits == 1
    assert cache_info.misses == 1

    slice_ = Slice(origin=(1, 0, 0), shape=shape2)

    mask_container.get(slice_)

    cache_info = mask_container._get_masks_for_slice[key].cache_info()
    assert cache_info.hits == 2
    assert cache_info.misses == 1


def test_for_datatile_1(masks):
    tile = DataTile(
        np.ones((1, 1, 1)),
        tile_slice=Slice(origin=(0, 0, 0), shape=Shape((1, 1, 1), sig_dims=2)),
        scheme_idx=0,
    )
    slice_ = masks.get_masks_for_slice(tile.tile_slice)
    assert slice_.shape == (1, 5)


def test_for_datatile_2(masks):
    tile = DataTile(
        np.ones((2 * 2, 10, 10)),
        tile_slice=Slice(origin=(0, 0, 0), shape=Shape((2 * 2, 10, 10), sig_dims=2)),
        scheme_idx=0,
    )
    slice_ = masks.get_masks_for_slice(tile.tile_slice)
    assert slice_.shape == (100, 5)


def test_for_datatile_with_scan_origin(masks):
    tile = DataTile(
        np.ones((2 * 2, 10, 10)),
        tile_slice=Slice(origin=(110, 0, 0), shape=Shape((2 * 2, 10, 10), sig_dims=2)),
        scheme_idx=0,
    )
    slice_ = masks.get_masks_for_slice(tile.tile_slice)
    assert slice_.shape == (100, 5)


def test_for_datatile_with_frame_origin(masks):
    tile = DataTile(
        np.ones((2 * 2, 1, 5)),
        tile_slice=Slice(origin=(110, 10, 10), shape=Shape((2 * 2, 1, 5), sig_dims=2)),
        scheme_idx=0,
    )
    slice_ = masks.get_masks_for_slice(tile.tile_slice)
    print(slice_)
    np.allclose(
        slice_,
        np.array([
            1, 0, 1, 0, 10,
            1, 0, 1, 0, 11,
            1, 0, 1, 0, 12,
            1, 0, 1, 0, 13,
            1, 0, 1, 0, 14,
        ]).reshape((5, 5))
    )


def test_merge_masks(masks):
    assert masks.computed_masks.shape == (5, 128, 128)


def test_sparse_pydata_cupy_default_unsupported():
    input_masks = [
        lambda: np.ones((128, 128)),
    ]
    factory = MaskContainer(
        mask_factories=input_masks,
        use_sparse=None,
        default_sparse='sparse.pydata',
        backend=CUPY,
    )
    assert factory.use_sparse is False


def test_raises_unknown_use_sparse():
    input_masks = [
        lambda: np.ones((128, 128)),
    ]
    with pytest.raises(ValueError):
        MaskContainer(
            mask_factories=input_masks,
            use_sparse='unknown',
        )



import typing
from typing import Union, Any
import pathlib
from functools import lru_cache
import importlib
import warnings
from typing_extensions import Literal
import numpy as np

from libertem.io.dataset.base import DataSetException, DataSet
from libertem.common.async_utils import sync_to_async
from libertem.common.scheduler import Scheduler


filetypes = {
    "hdf5": "libertem.io.dataset.hdf5.H5DataSet",
    "raw": "libertem.io.dataset.raw.RawFileDataSet",
    "raw_csr": "libertem.io.dataset.raw_csr.RawCSRDataSet",
    "mib": "libertem.io.dataset.mib.MIBDataSet",
    "blo": "libertem.io.dataset.blo.BloDataSet",
    "k2is": "libertem.io.dataset.k2is.K2ISDataSet",
    "ser": "libertem.io.dataset.ser.SERDataSet",
    "frms6": "libertem.io.dataset.frms6.FRMS6DataSet",
    "empad": "libertem.io.dataset.empad.EMPADDataSet",
    "dm": "libertem.io.dataset.dm.DMDataSet",
    "seq": "libertem.io.dataset.seq.SEQDataSet",
    "mrc": "libertem.io.dataset.mrc.MRCDataSet",
    "tvips": "libertem.io.dataset.tvips.TVIPSDataSet",
    "npy": "libertem.io.dataset.npy.NPYDataSet",
    "dask": "libertem.io.dataset.dask.DaskDataSet",
    "memory": "libertem.io.dataset.memory.MemoryDataSet",
}


@lru_cache
def build_extension_map() -> dict[str, list[str]]:
    ext_map = {}
    for typ_ in filetypes:
        cls = get_dataset_cls(typ_)
        for ext in cls.get_supported_extensions():
            try:
                ext_map[ext].append(typ_)
            except KeyError:
                ext_map[ext] = [typ_]
    return ext_map


@typing.overload
def _auto_load(
    path: str, enable_async: Literal[True], *args, executor, **kwargs,
) -> typing.Awaitable[DataSet]:
    ...


@typing.overload
def _auto_load(
    path: str, enable_async: Literal[False], *args, executor, **kwargs,
) -> DataSet:
    ...


@typing.overload
def _auto_load(
    path: str, enable_async: bool, *args, executor, **kwargs,
) -> typing.Union[DataSet, typing.Awaitable[DataSet]]:
    ...


def _auto_load(path, *args, executor, **kwargs):
    if path is None:
        raise DataSetException(
            "please specify the `path` argument to allow auto detection"
        )
    detected_params = detect(path, executor=executor)
    filetype_detected: typing.Optional[str] = detected_params.get('type', None)
    if filetype_detected is None:
        raise DataSetException(
            "could not determine DataSet type for file '%s'" % path,
        )
    return load(
        filetype_detected, path, *args, executor=executor, **kwargs
    )


@typing.overload
def load(
    filetype: str, *args, enable_async: Literal[True], executor, **kwargs,
) -> typing.Awaitable[DataSet]:
    ...


@typing.overload
def load(
    filetype: str, *args, enable_async: Literal[False], executor, **kwargs,
) -> DataSet:
    ...


@typing.overload
def load(
    filetype: str, *args, enable_async: bool, executor, **kwargs,
) -> typing.Union[DataSet, typing.Awaitable[DataSet]]:
    ...


def load(
    filetype: str,
    *args,
    enable_async: bool = False,
    executor,
    **kwargs,
):
    """
    Low-level method to load a dataset. Usually you will want
    to use Context.load instead!

    Parameters
    ----------
    filetype : str or DataSet type
        see libertem.io.dataset.filetypes for supported types, example: 'hdf5'

    executor : JobExecutor

    enable_async : bool
        If True, return a coroutine instead of blocking until the loading has
        finished.

    additional parameters are passed to the concrete DataSet implementation
    """
    if filetype == "auto":
        return _auto_load(*args, executor=executor, enable_async=enable_async, **kwargs)

    cls = get_dataset_cls(filetype)

    async def _init_async():
        ds = cls(*args, **kwargs)
        ds = await sync_to_async(ds.initialize, executor=executor.ensure_sync())
        workers = await executor.get_available_workers()
        scheduler = Scheduler(workers)
        # FIXME the partitioning should be dynamic
        # since the number of eligible workers may depend on
        # the set of UDFs that may or may not run on CPU or GPU
        # This is a workaround with a "best guess compromise"
        ds.set_num_cores(scheduler.effective_worker_count())
        await executor.run_function(ds.check_valid)
        return ds

    if enable_async:
        return _init_async()
    else:
        ds = cls(*args, **kwargs)
        ds = ds.initialize(executor)
        workers = executor.get_available_workers()
        scheduler = Scheduler(workers)
        ds.set_num_cores(scheduler.effective_worker_count())
        executor.run_function(ds.check_valid)
        return ds


def register_dataset_cls(filetype: str, cls: str) -> None:
    filetypes[filetype] = cls


def unregister_dataset_cls(filetype: str) -> None:
    del filetypes[filetype]


def get_dataset_cls(filetype: str) -> type[DataSet]:
    if not isinstance(filetype, str):
        return filetype
    try:
        ft = filetypes[filetype.lower()]
    except KeyError:
        raise DataSetException("unknown filetype: %s" % filetype)
    if not isinstance(ft, str):
        return ft
    parts = ft.split(".")
    module_name = ".".join(parts[:-1])
    cls_name = parts[-1]
    try:
        module = importlib.import_module(module_name)
    except ImportError as e:
        raise DataSetException("could not load dataset: %s" % str(e))
    cls: type[DataSet] = getattr(module, cls_name)
    return cls


def get_search_order(path: Union[str, np.ndarray]) -> list[str]:
    """
    Return the keys from filetypes in an order which
    is perhaps optimal for dataset auto-detection
    """
    extension_map = build_extension_map()
    search_order = list(filetypes.keys())
    try:
        # If the file format is registered, float the associated
        # datasets to the top of the search order (maintaining
        # the order in which they were first registered)
        file_format = pathlib.Path(path).suffix.strip().lstrip('.').lower()
        if file_format in extension_map:
            for ds_key in reversed(extension_map[file_format]):
                search_order.pop(search_order.index(ds_key))
                search_order = [ds_key] + search_order
    except (TypeError, ValueError):
        # Let downstream code handle the fact that
        # path cannot be cast to pathlib.Path or provide a suffix
        pass
    try:
        # If path has a shape attribute there is good chance
        # it implements the array interface and as such we should
        # check MemoryDataSet first
        _ = path.shape
        search_order.pop(search_order.index('memory'))
        search_order = ['memory'] + search_order
        warnings.warn('Auto-loading a MemoryDataSet is currently unsupported, '
                      'use ctx.load("memory", data=array).')
    except AttributeError:
        # Cannot interpret as a memory dataset
        pass
    return search_order


def detect(path: Union[str, np.ndarray], executor) -> dict[str, Any]:
    """
    Returns dataset's detected type, parameters and
    additional info.
    """
    search_order = get_search_order(path)
    for filetype in search_order:
        try:
            cls = get_dataset_cls(filetype)
            params = cls.detect_params(path, executor)
        except (NotImplementedError, DataSetException):
            continue
        if not params:
            continue
        params.update({"type": filetype})
        return params
    return {}


def get_extensions() -> set[str]:
    """
    Return supported extensions as a set of strings.

    Plain extensions only, no pattern!
    """
    types: set[str] = set()
    for filetype in filetypes.keys():
        cls = get_dataset_cls(filetype)
        types = types.union({ext.lower() for ext in cls.get_supported_extensions()})
    return types



import os
import pytest

from aio_utils import create_connection

pytestmark = [pytest.mark.web_api]


@pytest.mark.asyncio
async def test_browse_localfs(default_raw, base_url, http_client, local_cluster_url, default_token):
    await create_connection(base_url, http_client, local_cluster_url, default_token)
    browse_path = os.path.dirname(default_raw._path)
    raw_ds_filename = os.path.basename(default_raw._path)
    url = f"{base_url}/api/browse/localfs/"
    async with http_client.get(url, params={"path": browse_path, "token": default_token}) as resp:
        assert resp.status == 200
        listing = await resp.json()
        assert listing['status'] == 'ok'
        assert listing['messageType'] == 'DIRECTORY_LISTING'
        assert "drives" in listing
        assert "places" in listing
        assert "path" in listing
        assert "files" in listing
        assert "dirs" in listing
        assert listing["path"] == browse_path
        assert len(listing["files"]) >= 1
        defraw_found = False
        for entry in listing["files"]:
            assert set(entry.keys()) == {"name", "size", "ctime", "mtime", "owner"}
            if entry["name"] == raw_ds_filename:
                defraw_found = True
            assert defraw_found


@pytest.mark.asyncio
async def test_browse_localfs_stat(
    default_raw, base_url, http_client, local_cluster_url, default_token,
):
    await create_connection(base_url, http_client, local_cluster_url, default_token)
    browse_path = os.path.dirname(default_raw._path)
    raw_ds_filename = os.path.basename(default_raw._path)
    url = f"{base_url}/api/browse/localfs/stat/"

    # stat the directory:
    async with http_client.get(url, params={"path": browse_path, "token": default_token}) as resp:
        assert resp.status == 200
        stat_result = await resp.json()
        assert stat_result['status'] == 'ok'
        assert stat_result['messageType'] == 'STAT_RESULT'
        assert "basename" in stat_result
        assert stat_result["dirname"] == browse_path
        assert stat_result["path"] == browse_path
        for attr in ["size", "ctime", "mtime"]:
            assert attr in stat_result["stat"]
        assert stat_result["stat"]["isdir"]

    # stat the file:
    async with http_client.get(
        url, params={"path": default_raw._path, "token": default_token},
    ) as resp:
        assert resp.status == 200
        stat_result = await resp.json()
        assert stat_result['status'] == 'ok'
        assert stat_result['messageType'] == 'STAT_RESULT'
        assert stat_result["basename"] == raw_ds_filename
        assert stat_result["dirname"] == browse_path
        assert stat_result["path"] == default_raw._path
        for attr in ["size", "ctime", "mtime"]:
            assert attr in stat_result["stat"]
        assert not stat_result["stat"]["isdir"]
        assert stat_result["stat"]["isreg"]


@pytest.mark.asyncio
async def test_browse_localfs_stat_fail(
    default_raw, base_url, http_client, local_cluster_url, default_token,
):
    await create_connection(base_url, http_client, local_cluster_url, default_token)
    url = f"{base_url}/api/browse/localfs/stat/"
    path = '/does/not/really/exist/'

    # stat the directory:
    async with http_client.get(
        url, params={"path": path, "token": default_token}
    ) as resp:
        assert resp.status == 200
        stat_result = await resp.json()
        assert stat_result['status'] == 'error'
        assert stat_result['messageType'] == 'STAT_FAILED'
        assert 'alternative' in stat_result
        assert 'msg' in stat_result
        assert stat_result['path'] == path


@pytest.mark.asyncio
async def test_browse_localfs_fail(
    default_raw, base_url, http_client, local_cluster_url, default_token
):
    await create_connection(base_url, http_client, local_cluster_url, default_token)
    browse_path = os.path.join(
        os.path.dirname(default_raw._path),
        "does", "not", "exist"
    )
    url = f"{base_url}/api/browse/localfs/"
    async with http_client.get(
        url, params={"path": browse_path, "token": default_token}
    ) as resp:
        assert resp.status == 200
        listing = await resp.json()
        assert listing['status'] == 'error'
        assert listing['messageType'] == 'DIRECTORY_LISTING_FAILED'
